with Messages,Ada.Text_IO,Reps,Reps.IO;
use Messages;

package body Doubling is

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   
   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;

   function CompZ(F1,F2,Z: Taylor) return Taylor is
   begin
      return Comp1(Z,F1)+Comp2(Z,F2);
   end CompZ;
   
   
   function DCompZ(Rd: Radius; F1,F2,Z: Taylor) return Taylor is
   begin
      return Comp1(Z,Der1(Rd,F1))+Comp2(Z,Der2(Rd,F2));
   end DCompZ;
   
   
   --procedures for symmetric functions
   function CompZ(S,X,Y,Z: Taylor) return Taylor is
   begin
      return Comp(X,Z,S)+Comp(Y,Z,S);
   end CompZ;
   
   
   function DCompZ(Rd: Radius; S,X,Y,Z: Taylor) return Taylor is
      Der: Taylor;
   begin

      Der:=Der2(Rd,S);
      return Comp(X,Z,Der)+Comp(Y,Z,Der);
   end DCompZ;


   procedure FindZ(Epsi:  in out Radius; Rd: in Radius; S: in Taylor; Z: in out Taylor) is
      use Ada.Text_IO;
      use Reps.IO;
      use Tay_IO;
      ZZ,Sigma,H1,H2,H,H11,X,Y: Taylor;
      Eps,C1,C2,B,A0,B0,C0,Der:  Scalar;
      Co,Epss: Rep;
      N: Integer:=10;
      Rz: Radius:=Rho(Z);
      R: Radius:=Rho(S);
   begin
      if Trunc then
	 Epsi:=Rep(1.0E-19);
      end if;
      
      X:=Monom(Rz,1,0,Scal(One));
      Y:=Monom(Rz,0,1,Scal(One));
      
      B0:=Scal(-1.02668)*Scal(-1.96);
      A0:=Scal(-1.02668)*Scal(0.01);
      
      C0:=Scal(-1.0);
      
      Sigma:=S-Monom(R,0,1,B0)-Monom(R,0,2,A0)-Monom(R,0,0,C0);
      
      if  Trunc then
	 Epss:=One;
         while Epss>Epsi loop
	    
            ZZ:=Z;

            H11:=CompZ(Sigma,X,Y,ZZ)+Monom(Rz,0,0,Two*C0);
            H2:=DCompZ(Rd,Sigma,X,Y,ZZ);
            C1:=Norm(Rho(H11),H11);
            C2:=Norm(Rho(H2),H2);
            Der:=Inv( Sqrt( Sqr(B0)-Two*abs(A0)*C1 ) );
            Der:=Der*C2;
	    
            Z:=Inv(Two*A0)*(B0*Sqrt((-Two*A0/Sqr(B0))*H11)-Monom(Rz,0,0,B0) );

            H:=Z-ZZ;
            Epss:=MaxAbs(Norm(Rho(H),H));
	    
            New_Line;
            Put("Convergence in FindZ=");Txt_Put(Current_Output,Epss);
         end loop;
	 
      else

         B:=Ball(Zero,Epsi);
         Co:=Two;
         H11:=CompZ(Sigma,X,Y,Z)+Monom(Rz,0,0,Two*C0);

         H:=Inv(Two*A0)*(B0*Sqrt((-Two*A0/Sqr(B0))*H11)-Monom(Rz,0,0,B0) )-Z;

         Eps:=Norm(Rho(H),H);
         New_Line;
         Put("Eps in FindZ=");Show(Eps);

         while Co>One loop
            ZZ:=Z+Monom(Rz,0,0,B);
            H1:=CompZ(Sigma,X,Y,ZZ)+Monom(Rz,0,0,Two*C0);
            H2:=DCompZ(Rd,Sigma,X,Y,ZZ);

            C1:=Norm(Rho(H1),H1);
            C2:=Norm(Rho(H2),H2);

            Der:=Inv( Sqrt( Sqr(B0)-Two*abs(A0)*C1 ) );
            Der:=Der*C2;
	    if MaxAbs(Der)>One then
               New_Line;
               Message("Doubling.FindZ",Singular);
            end if;
	    Co:=MaxAbs( Eps/(MaxAbs(B)*(Scal(One) - Der )) );
	    B:=Rep(2.0)*B;
	    
         end loop;
         Z:=Z+Monom(Rz,0,0,Ball(Zero,MaxAbs(Eps/((Scal(One)-Der))) ));
      end if;

   end FindZ;


   procedure FindT(S,Z: in Taylor; Ce: in Scalar;  T: out Scalar) is

      SZ: Taylor;
      R: Radius;
      C1,C2: Scalar;
      C00,C01,C02,Co2,Co3,T1,T2,T3,T4: Scalar;
   begin

      SZ:=Comp1(Z,S);
      R:=Rho(SZ);

      C00:=ResetBall(Component(0,0,SZ));
      C01:=ResetBall(Component(0,1,SZ));
      C02:=ResetBall(Component(0,2,SZ));

      Co2:=Center(Component(0,2,SZ))+Disk(ResetCenter(Component(0,2,SZ)));
      Co3:=Center(Component(0,3,SZ))+Disk(ResetCenter(Component(0,3,SZ)));

      C1:=C00/(Sqr(Scal(R))*Scal(R))+C01/Sqr(Scal(R))+C02/Scal(R);
      C2:=Scal(Zero)/Sqr(Scal(R))+Scal(Zero)/Scal(R);
      T1:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      C1:=Scal(Zero)/(Sqr(Scal(R))*Scal(R))+C01/Sqr(Scal(R))+C02/Scal(R);
      C2:=C00/Sqr(Scal(R))+Scal(Zero)/Scal(R);
      T2:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      C1:=C00/(Sqr(Scal(R))*Scal(R))+Scal(Zero)/Sqr(Scal(R))+C02/Scal(R);
      C2:=Scal(Zero)/Sqr(Scal(R))+C01/Scal(R);
      T3:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      C1:=Scal(Zero)/(Sqr(Scal(R))*Scal(R))+Scal(Zero)/Sqr(Scal(R))+C02/Scal(R);
      C2:=C00/Sqr(Scal(R))+C01/Scal(R);
      T4:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      T:=T1;
      if    MaxAbs(Disk(ResetCenter(T2))) > MaxAbs(Disk(ResetCenter(T))) then
	 T:=T2;
      elsif MaxAbs(Disk(ResetCenter(T3))) > MaxAbs(Disk(ResetCenter(T))) then
	 T:=T3;
      elsif MaxAbs(Disk(ResetCenter(T4))) > MaxAbs(Disk(ResetCenter(T))) then
	 T:=T4;
      end if;

   end FindT;


   procedure DTh(Rd: in Radius; Ce,T: in Scalar; S,Z,H,DZ: in Taylor;  DT: out Scalar) is
      SZ,DSZ,HG: Taylor;
      C2,C3,C00,C01,C02,C22,Co2,Co3,DCo2,DCo3,DT1,DT2,DT3,DT4,DC3,DC2: Scalar;
      R1,R2: Radius;
   begin

      SZ:=Comp1(Z,S);
      DSZ:=Der1(Rd,S);
      DSZ:=Comp1(Z,DSZ);

      HG:=H;
      HG:=Comp1(Z,HG);

      DSZ:=DSZ*DZ+HG;

      R1:=Rho(SZ);
      R2:=Rho(DSZ);

      C00:=ResetBall(Component(0,0,SZ));
      C01:=ResetBall(Component(0,1,SZ));
      C02:=ResetBall(Component(0,2,SZ));

      Co2:=Center(Component(0,2,SZ))+Disk(ResetCenter(Component(0,2,SZ)));
      Co3:=Center(Component(0,3,SZ))+Disk(ResetCenter(Component(0,3,SZ)));
      DCo2:=Center(Component(0,2,DSZ))+Disk(ResetCenter(Component(0,2,DSZ)));
      DCo3:=Center(Component(0,3,DSZ))+Disk(ResetCenter(Component(0,3,DSZ)));

      DC2:=ResetBall(Component(0,0,DSZ))/Sqr(Scal(R2))+ResetBall(Component(0,1,DSZ))/Scal(R2);
      DC3:=ResetBall(Component(0,0,DSZ))/(R2*Sqr(Scal(R2)))+ResetBall(Component(0,1,DSZ))/Sqr(Scal(R2)) +ResetBall(Component(0,2,DSZ))/Scal(R2);
      C22:=C00/Sqr(Scal(R1))+C01/Scal(R1);

      C2:=Scal(Zero)/Sqr(Scal(R1))+Scal(Zero)/Scal(R1);
      C3:=C00/(Sqr(Scal(R1))*Scal(R1))+C01/Sqr(Scal(R1))+C02/Scal(R1);
      DT1:=Inv(Scal(-4))*(DCo3+DC3)/(Co2+C22)  +Inv(Scal(4))*(Co3+C3-Ce)/(Co2+C2)*(DCo2+DC2)/(Co2+C22);

      C2:=C00/Sqr(Scal(R1))+Scal(Zero)/Scal(R1);
      C3:=Scal(Zero)/(Sqr(Scal(R1))*Scal(R1))+C01/Sqr(Scal(R1))+C02/Scal(R1);
      DT2:=Inv(Scal(-4))*(DCo3+DC3)/(Co2+C22)  +Inv(Scal(4))*(Co3+C3-Ce)/(Co2+C2)*(DCo2+DC2)/(Co2+C22);

      C2:=Scal(Zero)/Sqr(Scal(R1))+C01/Scal(R1);
      C3:=C00/(Sqr(Scal(R1))*Scal(R1))+Scal(Zero)/Sqr(Scal(R1))+C02/Scal(R1);
      DT3:=Inv(Scal(-4))*(DCo3+DC3)/(Co2+C22)  +Inv(Scal(4))*(Co3+C3-Ce)/(Co2+C2)*(DCo2+DC2)/(Co2+C22);

      C2:=C00/Sqr(Scal(R1))+C01/Scal(R1);
      C3:=Scal(Zero)/(Sqr(Scal(R1))*Scal(R1))+Scal(Zero)/Sqr(Scal(R1))+C02/Scal(R1);
      DT4:=Inv(Scal(-4))*(DCo3+DC3)/(Co2+C22)  +Inv(Scal(4))*(Co3+C3-Ce)/(Co2+C2)*(DCo2+DC2)/(Co2+C22);



      DT:=DT1;
      if    MaxAbs(Disk(ResetCenter(DT2))) > MaxAbs(Disk(ResetCenter(DT))) then
	 DT:=DT2;
      elsif MaxAbs(Disk(ResetCenter(DT3))) > MaxAbs(Disk(ResetCenter(DT))) then
	 DT:=DT3;
      elsif MaxAbs(Disk(ResetCenter(DT4))) > MaxAbs(Disk(ResetCenter(DT))) then
	 DT:=DT4;
      end if;
      
   end DTh;


   procedure FindLambda(Rdz: in Radius; Z,S0,P: in Taylor; Lambda: out Scalar) is
      Iter: constant Integer := 200;
      Eps: constant Radius := 1.0E-19;
      B,Epss,C,M: Scalar;
      Co,Crit: Rep;
      P1: Taylor;
   begin

      Lambda:=Scal(-0.2488);
      P1:=Der1(Rdz,P);
      Crit:=One;
      for I in 1 .. Iter loop
         Lambda:=Center(Lambda);
         C:=Val(Lambda,Scal(Zero),P)/Val(Lambda,Scal(Zero),P1);
	 Lambda:=Lambda-C;
         Crit:=MaxAbs(C);
         exit when MaxAbs(C)<Eps;
      end loop;
      Lambda:=Center(Lambda);

      if Trunc then return; end if;
      --Contraction Mapping
      B:=Ball(Rep(1.0E-19));
      M:=-Inv(Val(Lambda,Scal(Zero),P1));
      Co:=Two;
      Epss:=Val(Lambda,Scal(Zero),Z);
      Epss:=Val(Epss,Scal(Zero),S0);
      while Co>One loop
	 Co:=MaxAbs(Epss / ( MaxAbs(B)*(Val(Lambda+M*B,Scal(Zero),P1)*M) ) );
         B:=Rep(1.2)*B;
      end loop;
      Lambda:=Lambda+M*B;
   end FindLambda;


   procedure RG(Eps,Rd,Rdz: in Radius; Ce: in Scalar;  Lambda, Mu,T: out Scalar; S,Z: in out Taylor  ) is
      use Ada.Text_IO, Reps.IO;
      X,Y,S0: Taylor;
      R, Rdp: Radius;
   begin
      S0:=S;
      PG(Eps,Rd,S,Z,Ce,T);
      FindLambda(Rdz,Z,S0,S,Lambda);
      Rdp:=Inf(abs(Scal(Rd)/Lambda));
      R:=Inf(Widen(abs(Scal(Rho(S))/Lambda)));
      X:=Monom(R,1,0,Lambda);
      Y:=Monom(R,0,1,Lambda);
      S:=Comp(X,Y,S);
      if Trunc then
         Mu:=Val(Scal(One),Scal(Zero),Der1(Rep(1.1),S)) ;--any radius larger than 1 is OK
      else
         Mu:=Val(Scal(One),Scal(Zero),Der1(Sup(Widen(Scal(One))),S)) ;--any radius larger than 1 is OK
      end if;
      S:=Inv(Mu)*S;
      S:=S-Monom(R,0,0,Val(Scal(One),Scal(Zero),S));
   end RG;


   procedure PG(Eps,Rd: in Radius;  S,Z: in out Taylor;  Ce: in Scalar; T: out Scalar) is
      use Ada.Text_IO, Reps.IO;
      X,Y: Taylor;
      R: constant Radius:=Rho(Z);
      Epps,Rm,Rx: Radius;
      C1,C2,C00,C01,C02,Co2,Co3,T1,T2,T3,T4: Scalar;
      R1: Radius;
   begin
      Epps:=Eps;

      if Trunc then
         FindZ(Epps,Rd,S,Z);
      end if;

      S:=Comp1(Z,S);
      R1:=Rho(S);
      
      C00:=ResetBall(Component(0,0,S));
      C01:=ResetBall(Component(0,1,S));
      C02:=ResetBall(Component(0,2,S));

      Co2:=Center(Component(0,2,S))+Disk(ResetCenter(Component(0,2,S)));
      Co3:=Center(Component(0,3,S))+Disk(ResetCenter(Component(0,3,S)));

      C1:=C00/(Sqr(Scal(R))*Scal(R))+C01/Sqr(Scal(R))+C02/Scal(R);
      C2:=Scal(Zero)/Sqr(Scal(R))+Scal(Zero)/Scal(R);
      T1:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      C1:=Scal(Zero)/(Sqr(Scal(R))*Scal(R))+C01/Sqr(Scal(R))+C02/Scal(R);
      C2:=C00/Sqr(Scal(R))+Scal(Zero)/Scal(R);
      T2:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      C1:=C00/(Sqr(Scal(R))*Scal(R))+Scal(Zero)/Sqr(Scal(R))+C02/Scal(R);
      C2:=Scal(Zero)/Sqr(Scal(R))+C01/Scal(R);
      T3:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      C1:=Scal(Zero)/(Sqr(Scal(R))*Scal(R))+Scal(Zero)/Sqr(Scal(R))+C02/Scal(R);
      C2:=C00/Sqr(Scal(R))+C01/Scal(R);
      T4:=Inv(Scal(4))*(Ce-(Co3+C1))/(Co2+C2);

      T:=T1;
      if    MaxAbs(Disk(ResetCenter(T2))) > MaxAbs(Disk(ResetCenter(T))) then
	 T:=T2;
      elsif MaxAbs(Disk(ResetCenter(T3))) > MaxAbs(Disk(ResetCenter(T))) then
	 T:=T3;
      elsif MaxAbs(Disk(ResetCenter(T4))) > MaxAbs(Disk(ResetCenter(T))) then
	 T:=T4;
      end if;
      Rm:=Rho(S);
      Rx:= Inf(Widen(Rm*(Scal(One)-Rm*abs(Widen(T)))));
      X:=Monom(Rx,1,0,Scal(One))+Monom(Rx,2,0,T);
      Y:=Monom(Rx,0,1,Scal(One))+Monom(Rx,0,2,T);
      S:=(Monom(R,0,0,Scal(One))+Monom(R,0,1,Two*T))*Comp(X,Y,S);
   end PG;


   procedure DLambdah(Rp1: in Radius; Lambda: in Scalar; P,DP: in Taylor; DL: out Scalar)  is
      P1: Taylor;
   begin
      P1:=Der1(Rp1,P);
      DL:=-Val(Lambda,Scal(Zero),DP)/Val(Lambda,Scal(Zero),P1);
   end DLambdah;


   function DLambdah(Rp1: Radius; Lambda: Scalar; P,DP: Taylor) return Scalar is
      DL: Scalar;
   begin
      DLambdah(Rp1,Lambda,P,DP,DL);
      return DL;
   end DLambdah;
   
   
   procedure DMuh(Rdp1,Rp1,Rp11: in Radius; Lambda,DL: in Scalar; P,DP: in  Taylor; DM: out Scalar) is
      use Ada.Text_IO, Reps.IO;
      DP1,P11,P1: Taylor;
      ML: constant Scalar:=abs(Lambda);
   begin
      DP1:=Der1(Rdp1,DP);
      P1:=Der1(Rp1,P);
      P11:=Der1(Rp11,P1);
      DM:=Lambda*Val(Lambda,ScalZero,DP1)+Lambda*Val(Lambda,ScalZero,P11)*DL+DL*Val(Lambda,ScalZero,P1);
   end DMuh;


   function DMuh(Rdp1,Rp1,Rp11: Radius; Lambda,DL: Scalar; P,DP: Taylor) return Scalar is
      DM: Scalar;
   begin
      DMuh(Rdp1,Rp1,Rp11,Lambda,DL,P,DP,DM);
      return DM;
   end DMuh;


   procedure DZh(Rd: in Radius; S,Z: in Taylor; H: in out Taylor) is
      X,Y: Taylor;
      R: constant Radius:=Rho(S);
      Rz: constant Radius:=Rho(Z);
   begin
      X:=Monom(Rz,1,0,Scal(One));
      Y:=Monom(Rz,0,1,Scal(One));
      H:=-( Comp(X,Z,H)+Comp(Y,Z,H)) / ( Comp(X,Z, Der2(Rd,S) ) + Comp(Y,Z,Der2(Rd,S)) );
   end DZh;


   procedure DPGh(Rd,Rdz: in Radius; Ce,T: in Scalar; S: in Taylor; Z: in Taylor; H: in out Taylor) is
      use Ada.Text_IO, Reps.IO;
      SGE,S1GE,S2GE,DZ,M,HGE,X,Y: Taylor;
      DT: Scalar;
      R: constant Radius:=Rho(S);
      Rz: constant Radius:=Rho(Z);
      Rx,Rm: Radius;
   begin
      Ders(Rd,S,S1GE,S2GE);
      S1GE:=Comp1(Z,S1GE);
      S2GE:=Comp1(Z,S2GE);
      HGE:=H;
      HGE:=Comp1(Z,HGE);
      SGE:=Comp1(Z,S);
      M:=Monom(R,0,0,Scal(One))+Monom(R,0,1,Two*T);
      Rm:=Rho(S1GE);
      Rx:= Inf(Widen(Rm*(Scal(One)-Rm*abs(T))));
      X:=Monom(Rx,1,0,Scal(One))+Monom(Rx,2,0,T);
      Y:=Monom(Rx,0,1,Scal(One))+Monom(Rx,0,2,T);
      S1GE:=Comp(X,Y,S1GE);
      S2GE:=Comp(X,Y,S2GE);
      Rm:=Rho(SGE);
      Rx:= Inf(Widen(Rm*(Scal(One)-Rm*abs(T))));
      SetRho(Rx,X);
      SetRho(Rx,Y);
      SGE:=Comp(X,Y,SGE);
      Rm:=Rho(HGE);
      Rx:= Inf( Widen(Rm*(Scal(One)-Rm*abs(T))) );
      SetRho(Rx,X);
      SetRho(Rx,Y);
      HGE:=Comp(X,Y,HGE);
      DZ:=H;
      DZh(Rd,S,Z,DZ);
      DTh(Rd,Ce,T,S,Z,H,DZ,DT);
      H:=Monom(R,0,1,Two*DT)*SGE;
      H:=H+M*S1GE*Comp(X,Y,DZ);
      Rm:=Rdz;
      Rx:= Inf(Widen(Rm*(Scal(One)-Rm*abs(T))));
      SetRho(Rx,X);
      SetRho(Rx,Y);
      H:=H+M*S1GE*(Comp(X,Y,Der1(Rdz,Z))*Monom(R,2,0,DT)+Comp(X,Y,Der2(Rdz,Z))*Monom(R,0,2,DT));
      H:=H+M*S2GE*Monom(R,0,2,DT)+M*HGE;
   end DPGh;


   procedure DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11: in Radius; Ce,Lambda,Mu: in Scalar; S: in Taylor; Z: in Taylor; H: in out Taylor) is
      use Ada.Text_IO, Reps.IO;
      ZZ,P,DP,X,Y,P1,P2,XX,YY,X1,Y1: Taylor;
      T,DL,DM: Scalar;
      R: constant Radius:=Rho(S);
      Rp1l,Rp2l,Rp12l,Rpl,Rdpl: Radius;
   begin
      P:=S;
      DP:=H;
      ZZ:=Z;
      PG(Eps,Rd,P,ZZ,Ce,T);--finds Z,P,T
      DPGh(Rd,Rdz,Ce,T,S,ZZ,DP);
      DLambdah(Rp1,Lambda,P,DP,DL);
      DMuh(Rdp1,Rp1,Rp11,Lambda,DL,P,DP,DM);
      Rpl:=Inf(Rep(0.97)*Scal(Rho(P))/abs(Lambda));
      X:=Monom(Rpl,1,0,Lambda);
      Y:=Monom(Rpl,0,1,Lambda);
      Rdpl:=Inf(Rep(0.97)*Scal(Rho(DP))/abs(Lambda));
      X1:=Monom(Rdpl,1,0,Lambda);
      Y1:=Monom(Rdpl,0,1,Lambda);
      Ders(Rp1,P,P1,P2);
      Rp1l:=Inf(Rep(0.97)*Scal(Rho(P1))/abs(Lambda));
      Rp2l:=Inf(Rep(0.97)*Scal(Rho(P2))/abs(Lambda));
      Rp12l:=Rep'Min(Rp1l,Rp2l);
      
      XX:=Monom(Rp12l,1,0,Lambda);
      YY:=Monom(Rp12l,0,1,Lambda);
      H:=Inv(Mu)*( (-DM)*Inv(Mu)*Comp(X,Y,P)+Comp(X1,Y1,DP)+DL*Monom(Rp12l,1,0,Scal(One))*Comp(XX,YY,P1) + DL*Monom(Rp12l,0,1,Scal(One))*Comp(XX,YY,P2)  );
      H:=H-Monom(Rho(H),0,0,Val(Scal(One),Scal(Zero),H));
   end DRGh;


   function DD(N: Integer) return Integer is
      K: Integer;
   begin
      K:=0;
      for I in -1..(N-1) loop
         for J in IMax(0,I)..(N-1-I) loop
            K:=K+1;
         end loop;
      end loop;
      return K;
   end DD;
   

   procedure PrepNewt(N: in Integer; Ce: in Scalar; Eps,Rd,Rdz,Rdp1,Rp1,Rp11: in Radius;  S,Z: in Taylor;  D: out Matrix; V:   out Vector) is
      use Ada.Text_IO, Reps.IO;
      Dim: constant Integer := DD(N);
      K: Integer;
      Mu,Lambda,T: Scalar;
      F: Taylor;
      Nrm: Vector(1..Dim);
      H,Clmn: TayVector(1..Dim);
      ZZ: Taylor;
      R: constant Radius:=Rho(S);
   begin
      --Normalized basis vectors (H) and matrix columns (Clmn)
      --Right hand side vector of the linear algebra in Newton (V)
      F:=S;
      ZZ:=Z;
      RG(Eps,Rd,Rdz,Ce,Lambda,Mu,T,F,ZZ);
      F:=F-S;
      K:=1;
      for I in -1..(N-1) loop
         for J in IMax(0,I)..(N-1-I) loop
            if I=-1 or else I=J then
               H(K):=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
            else
               H(K):=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
            end if;

            Nrm(K):=Norm(R,H(K));
            H(K):=Inv(Nrm(K))*H(K); --normalized vector (xi+eta)
            Clmn(K):=H(K);
            ZZ:=Z;
	    DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,ZZ,Clmn(K));
	    Clmn(K):=Clmn(K)-H(K);
            V(K):=(-Nrm(K))*Component(I+1,J,F);
            F:=F+V(K)*H(K);
            K:=K+1;
         end loop;
      end loop;

      --Components of columns along the basis vectors
      for M in Clmn'Range loop
         K:=Dim;
         for I in reverse -1..(N-1) loop
            for J in reverse IMax(0,I)..(N-1-I) loop
               D(K,M):=Nrm(K)*Component(I+1,J,Clmn(M));
               Clmn(M):=Clmn(M)-D(K,M)*H(K);
               K:=K-1;
            end loop;
         end loop;
      end loop;
   end PrepNewt;


   procedure FindS(Eps: in Radius; N: in Integer; Ce: in Scalar; Rd,Rdz,Rdp1,Rp1,Rp11: in Radius;  S,Z: in out Taylor  ) is
      Dim: constant Integer := DD(N);
      Iter: constant Integer := 50;
      K: Integer;
      Cond: Scalar;
      ZZ,DS: Taylor;
      Pvt: Pivot(1..Dim);
      V,Work: Vector(1..Dim);
      D: Matrix(1..Dim,1..Dim);
      H: TayVector(1..Dim); --2N is the number of symmetric basis vectors of order <=N
      R: constant Radius:=Rho(S);
   begin

      --Normalized basis vectors
      K:=1;
      for I in -1..(N-1) loop
         for J in IMax(0,I)..(N-1-I) loop
            if I=-1 or else I=J  then
               H(K):=Monom(R,I+1,J,ScalOne);
            else
               H(K):=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne);
            end if;
            H(K):=Inv(Norm(R,H(K)))*H(K);
            K:=K+1;
         end loop;
      end loop;

      --Newton for F(DS):=P[R(S+DS)-(S+DS)]=0. Numerical for now
      New_Line;
      for I in 1..Iter loop
         ZZ:=Z;
         PrepNewt(N,Ce,Eps,Rd,Rdz,Rdp1,Rp1,Rp11,S,ZZ,D,V);
         Decomp(D,Cond,Pvt,Work);
         Solve(D,V,Pvt);
         SetZero(R,DS);
         for J in H'Range loop
            DS:=DS+V(J)*H(J);
         end loop;
         S:=S+DS;
         exit when MaxAbs(Norm(R,DS))<Eps;
         Show("Convergence criterion = ",MaxAbs(Norm(R,DS)));
      end loop;
   end FindS;


   function Ind(II: in Integer) return String is
      Num: String:="000";
   begin
      if II=1 then Num:="001";
      elsif II=2 then Num:="002";
      elsif II=3 then Num:="003";
      elsif II=4 then Num:="004";
      elsif II=5 then Num:="005";
      elsif II=6 then Num:="006";
      elsif II=7 then Num:="007";
      elsif II=8 then Num:="008";
      elsif II=9 then Num:="009";

      elsif II=10 then Num:="010";
      elsif II=11 then Num:="011";
      elsif II=12 then Num:="012";
      elsif II=13 then Num:="013";
      elsif II=14 then Num:="014";
      elsif II=15 then Num:="015";
      elsif II=16 then Num:="016";
      elsif II=17 then Num:="017";
      elsif II=18 then Num:="018";
      elsif II=19 then Num:="019";

      elsif II=20 then Num:="020";
      elsif II=21 then Num:="021";
      elsif II=22 then Num:="022";
      elsif II=23 then Num:="023";
      elsif II=24 then Num:="024";
      elsif II=25 then Num:="025";
      elsif II=26 then Num:="026";
      elsif II=27 then Num:="027";
      elsif II=28 then Num:="028";
      elsif II=29 then Num:="029";

      elsif II=30 then Num:="030";
      elsif II=31 then Num:="031";
      elsif II=32 then Num:="032";
      elsif II=33 then Num:="033";
      elsif II=34 then Num:="034";
      elsif II=35 then Num:="035";
      elsif II=36 then Num:="036";
      elsif II=37 then Num:="037";
      elsif II=38 then Num:="038";
      elsif II=39 then Num:="039";

      elsif II=40 then Num:="040";
      elsif II=41 then Num:="041";
      elsif II=42 then Num:="042";
      elsif II=43 then Num:="043";
      elsif II=44 then Num:="044";
      elsif II=45 then Num:="045";
      elsif II=46 then Num:="046";
      elsif II=47 then Num:="047";
      elsif II=48 then Num:="048";
      elsif II=49 then Num:="049";

      elsif II=50 then Num:="050";
      elsif II=51 then Num:="051";
      elsif II=52 then Num:="052";
      elsif II=53 then Num:="053";
      elsif II=54 then Num:="054";
      elsif II=55 then Num:="055";
      elsif II=56 then Num:="056";
      elsif II=57 then Num:="057";
      elsif II=58 then Num:="058";
      elsif II=59 then Num:="059";

      elsif II=60 then Num:="060";
      elsif II=61 then Num:="061";
      elsif II=62 then Num:="062";
      elsif II=63 then Num:="063";
      elsif II=64 then Num:="064";
      elsif II=65 then Num:="065";
      elsif II=66 then Num:="066";
      elsif II=67 then Num:="067";
      elsif II=68 then Num:="068";
      elsif II=69 then Num:="069";

      elsif II=70 then Num:="070";
      elsif II=71 then Num:="071";
      elsif II=72 then Num:="072";
      elsif II=73 then Num:="073";
      elsif II=74 then Num:="074";
      elsif II=75 then Num:="075";
      elsif II=76 then Num:="076";
      elsif II=77 then Num:="077";
      elsif II=78 then Num:="078";
      elsif II=79 then Num:="079";

      elsif II=80 then Num:="080";
      elsif II=81 then Num:="081";
      elsif II=82 then Num:="082";
      elsif II=83 then Num:="083";
      elsif II=84 then Num:="084";
      elsif II=85 then Num:="085";
      elsif II=86 then Num:="086";
      elsif II=87 then Num:="087";
      elsif II=88 then Num:="088";
      elsif II=89 then Num:="089";

      elsif II=90 then Num:="090";
      elsif II=91 then Num:="091";
      elsif II=92 then Num:="092";
      elsif II=93 then Num:="093";
      elsif II=94 then Num:="094";
      elsif II=95 then Num:="095";
      elsif II=96 then Num:="096";
      elsif II=97 then Num:="097";
      elsif II=98 then Num:="098";
      elsif II=99 then Num:="099";

      elsif II=100 then Num:="100";
      elsif II=101 then Num:="101";
      elsif II=102 then Num:="102";
      elsif II=103 then Num:="103";
      elsif II=104 then Num:="104";
      elsif II=105 then Num:="105";
      elsif II=106 then Num:="106";
      elsif II=107 then Num:="107";
      elsif II=108 then Num:="108";
      elsif II=109 then Num:="109";


      elsif II=110 then Num:="110";
      elsif II=111 then Num:="111";
      elsif II=112 then Num:="112";
      elsif II=113 then Num:="113";
      elsif II=114 then Num:="114";
      elsif II=115 then Num:="115";
      elsif II=116 then Num:="116";
      elsif II=117 then Num:="117";
      elsif II=118 then Num:="118";
      elsif II=119 then Num:="119";


      elsif II=120 then Num:="120";
      elsif II=121 then Num:="121";
      elsif II=122 then Num:="122";
      elsif II=123 then Num:="123";
      elsif II=124 then Num:="124";
      elsif II=125 then Num:="125";
      elsif II=126 then Num:="126";
      elsif II=127 then Num:="127";
      elsif II=128 then Num:="128";
      elsif II=129 then Num:="129";


      elsif II=130 then Num:="130";
      elsif II=131 then Num:="131";
      elsif II=132 then Num:="132";
      elsif II=133 then Num:="133";
      elsif II=134 then Num:="134";
      elsif II=135 then Num:="135";
      elsif II=136 then Num:="136";
      elsif II=137 then Num:="137";
      elsif II=138 then Num:="138";
      elsif II=139 then Num:="139";


      elsif II=140 then Num:="140";
      elsif II=141 then Num:="141";
      elsif II=142 then Num:="142";
      elsif II=143 then Num:="143";
      elsif II=144 then Num:="144";
      elsif II=145 then Num:="145";
      elsif II=146 then Num:="146";
      elsif II=147 then Num:="147";
      elsif II=148 then Num:="148";
      elsif II=149 then Num:="149";


      elsif II=150 then Num:="150";
      elsif II=151 then Num:="151";
      elsif II=152 then Num:="152";
      elsif II=153 then Num:="153";
      elsif II=154 then Num:="154";
      elsif II=155 then Num:="155";
      elsif II=156 then Num:="156";
      elsif II=157 then Num:="157";
      elsif II=158 then Num:="158";
      elsif II=159 then Num:="159";


      elsif II=160 then Num:="160";
      elsif II=161 then Num:="161";
      elsif II=162 then Num:="162";
      elsif II=163 then Num:="163";
      elsif II=164 then Num:="164";
      elsif II=165 then Num:="165";
      elsif II=166 then Num:="166";
      elsif II=167 then Num:="167";
      elsif II=168 then Num:="168";
      elsif II=169 then Num:="169";


      elsif II=170 then Num:="170";
      elsif II=171 then Num:="171";
      elsif II=172 then Num:="172";
      elsif II=173 then Num:="173";
      elsif II=174 then Num:="174";
      elsif II=175 then Num:="175";
      elsif II=176 then Num:="176";
      elsif II=177 then Num:="177";
      elsif II=178 then Num:="178";
      elsif II=179 then Num:="179";


      elsif II=180 then Num:="180";
      elsif II=181 then Num:="181";
      elsif II=182 then Num:="182";
      elsif II=183 then Num:="183";
      elsif II=184 then Num:="184";
      elsif II=185 then Num:="185";
      elsif II=186 then Num:="186";
      elsif II=187 then Num:="187";
      elsif II=188 then Num:="188";
      elsif II=189 then Num:="189";


      elsif II=190 then Num:="190";
      elsif II=191 then Num:="191";
      elsif II=192 then Num:="192";
      elsif II=193 then Num:="193";
      elsif II=194 then Num:="194";
      elsif II=195 then Num:="195";
      elsif II=196 then Num:="196";
      elsif II=197 then Num:="197";
      elsif II=198 then Num:="198";
      elsif II=199 then Num:="199";


      elsif II=200 then Num:="200";
      elsif II=201 then Num:="201";
      elsif II=202 then Num:="202";
      elsif II=203 then Num:="203";
      elsif II=204 then Num:="204";
      elsif II=205 then Num:="205";
      elsif II=206 then Num:="206";
      elsif II=207 then Num:="207";
      elsif II=208 then Num:="208";
      elsif II=209 then Num:="209";

      elsif II=210 then Num:="210";
      elsif II=211 then Num:="211";
      elsif II=212 then Num:="212";
      elsif II=213 then Num:="213";
      elsif II=214 then Num:="214";
      elsif II=215 then Num:="215";
      elsif II=216 then Num:="216";
      elsif II=217 then Num:="217";
      elsif II=218 then Num:="218";
      elsif II=219 then Num:="219";

      elsif II=220 then Num:="220";
      elsif II=221 then Num:="221";
      elsif II=222 then Num:="222";
      elsif II=223 then Num:="223";
      elsif II=224 then Num:="224";
      elsif II=225 then Num:="225";
      elsif II=226 then Num:="226";
      elsif II=227 then Num:="227";
      elsif II=228 then Num:="228";
      elsif II=229 then Num:="229";

      elsif II=230 then Num:="230";
      elsif II=231 then Num:="231";
      elsif II=232 then Num:="232";
      elsif II=233 then Num:="233";
      elsif II=234 then Num:="234";
      elsif II=235 then Num:="235";
      elsif II=236 then Num:="236";
      elsif II=237 then Num:="237";
      elsif II=238 then Num:="238";
      elsif II=239 then Num:="239";

      elsif II=240 then Num:="240";
      elsif II=241 then Num:="241";
      elsif II=242 then Num:="242";
      elsif II=243 then Num:="243";
      elsif II=244 then Num:="244";
      elsif II=245 then Num:="245";
      elsif II=246 then Num:="246";
      elsif II=247 then Num:="247";
      elsif II=248 then Num:="248";
      elsif II=249 then Num:="249";

      elsif II=250 then Num:="250";
      elsif II=251 then Num:="251";
      elsif II=252 then Num:="252";
      elsif II=253 then Num:="253";
      elsif II=254 then Num:="254";
      elsif II=255 then Num:="255";
      elsif II=256 then Num:="256";
      elsif II=257 then Num:="257";
      elsif II=258 then Num:="258";
      elsif II=259 then Num:="259";

      elsif II=260 then Num:="260";
      elsif II=261 then Num:="261";
      elsif II=262 then Num:="262";
      elsif II=263 then Num:="263";
      elsif II=264 then Num:="264";
      elsif II=265 then Num:="265";
      elsif II=266 then Num:="266";
      elsif II=267 then Num:="267";
      elsif II=268 then Num:="268";
      elsif II=269 then Num:="269";

      elsif II=270 then Num:="270";
      elsif II=271 then Num:="271";
      elsif II=272 then Num:="272";
      elsif II=273 then Num:="273";
      elsif II=274 then Num:="274";
      elsif II=275 then Num:="275";
      elsif II=276 then Num:="276";
      elsif II=277 then Num:="277";
      elsif II=278 then Num:="278";
      elsif II=279 then Num:="279";

      else New_Line;
      Num:="000";
      Put("Dimension too large...");
      end if;

      return Num;
   end Ind;


   procedure DRGG(N: in Integer; Eps,Rd,Rdz,Rdp1,Rp1,Rp11: in Radius; Ce,Lambda,Mu: in Scalar; S,Z: in Taylor; D: out Matrix) is
      use Ada.Text_IO, Reps.IO;
      --     Dim: constant Integer := N*N;
      Dim: constant Integer := DD(N);
      K: Integer;
      Nrm: Vector(1..Dim);
      H: TayVector(1..Dim);
      HH: Taylor;
      Nrm1: Scalar;
      R:  constant Radius:=Rho(S);
   begin
      K:=1;
      --Normalized basis vectors (H) and matrix columns (Clmn)
      --Right hand side vector of the linear algebra in Newton (V)

      for I in -1..(N-1) loop
         for J in IMax(0,I)..(N-1-I) loop
            if I=-1 or else I=J then
               H(K):=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
            else
               H(K):=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
            end if;

            Nrm(K):=Norm(R,H(K));
            H(K):=Inv(Nrm(K))*H(K); --normalized vector (xi+eta)
            DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,H(K));
            K:=K+1;
         end loop;
      end loop;


      for M in H'Range loop
         K:=Dim;
         for I in reverse -1..(N-1) loop
            for J in reverse IMax(0,I)..(N-1-I) loop
               if I=-1 or else I=J then
                  HH:=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
               else
                  HH:=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
               end if;

               Nrm1:=Norm(R,HH);
               HH:=Inv(Nrm1)*HH;
               D(K,M):=Nrm1*Component(I+1,J,H(M));
               H(M):=H(M)-D(K,M)*HH;

               K:=K-1;

            end loop;
         end loop;
      end loop;
   end DRGG;


   function PiProj(R: in Radius; N,II: in Integer; H: in Taylor) return Scalar is
      Dim: constant Integer := DD(N);
      HH,G: Taylor;
      Co: Vector(1..Dim);
      K: Integer;
      CC,Nrm1: Scalar;
   begin
      G:=H;
      K:=Dim;
      for I in reverse -1..(N-1) loop
         for J in reverse IMax(0,I)..(N-1-I) loop
            if I=-1 or else I=J then
               HH:=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
            else
               HH:=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
            end if;

            Nrm1:=Norm(R,HH);
            HH:=Inv(Nrm1)*HH;

            Co(K):=Nrm1*Component(I+1,J,G);
            G:=G-Co(K)*HH;

            CC:=Co(K);
            if K=II then exit; end if;

            K:=K-1;
         end loop;
      end loop;

      return CC;
   end PiProj;


   function PProj(R: in Radius; D,N,II: in Integer; Eps: in Scalar; H: in Taylor; B: in Matrix) return Scalar is
      use Ada.Text_IO;
      use Reps.IO;
      Dim: constant Integer := DD(N);
      HH,G: Taylor;
      V,Co: Vector(1..D);
      K: Integer;
      CC,MaxCo,Nrm1: Scalar;
   begin

      G:=H;
      K:=Dim;
      MaxCo:=Scal(Zero);
      for I in reverse -1..(N-1) loop
         for J in reverse IMax(0,I)..(N-1-I) loop
            if I=-1  or else I=J then
               HH:=Monom(R,I+1,J,Scal(One)); --vector (xi+eta)
            else
               HH:=Monom(R,I+1,J,Scal(One))+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,Scal(One)); --vector (xi+eta)
            end if;

            Nrm1:=Abs(Norm(R,HH));
            CC:=Component(I+1,J,G);
            G:=G-CC*HH;
            CC:=Nrm1*CC;

            if Sup(abs(CC))>Sup(MaxCo) then
               MaxCo:=abs(CC);
            end if;
            if K<=D then  Co(K):=CC;
            end if;

            K:=K-1;

         end loop;
      end loop;

      V:=B*Co;
      return V(II)+Eps*MaxCo;
   end PProj;


   function PProj(R: in Radius; D,N: in Integer; EpsOne,EpsInf: in Scalar; H: in Taylor; B: in Matrix) return Vector is
      use Ada.Text_IO;
      use Reps.IO;
      Dim: constant Integer := DD(N);
      HH,G: Taylor;
      V,Co: Vector(1..D);
      VV: Vector(1..D+1);
      K: Integer;
      CCC,CC,Nrm1,MaxCo,MaxCoo: Scalar;
   begin

      G:=H;
      K:=Dim;
      MaxCo:=Scal(Zero);
      MaxCoo:=Scal(Zero);
      for I in reverse -1..(N-1) loop
         for J in reverse IMax(0,I)..(N-1-I) loop
            if I=-1 or else I=J then
               HH:=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
               Nrm1:=Norm(R,HH);
               CC:=Nrm1*Component(I+1,J,G);
               ZeroComponent(I+1,J,G);
            else
               HH:=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
               Nrm1:=Norm(R,HH);
               CCC:=Component(I+1,J,G);
               CC:=Nrm1*CCC;
               ZeroComponent(I+1,J,G);
               G:=G-CCC*Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne);
	    end if;

            if Sup(abs(CC))>Sup(MaxCoo) then
               MaxCoo:=abs(CC);
            end if;
            if K<=D then  Co(K):=CC;
            MaxCo:=MaxCo+abs(Co(K));
            end if;

            K:=K-1;
         end loop;
      end loop;

      V:=B*Co;
      for I in V'Range loop
         VV(I):=V(I);
         if I=1 then
            VV(I):=VV(I)+Ball(Sup(EpsInf*MaxCoo));
         end if;
         if I=2 then
            VV(I):=VV(I)+Ball(Sup(EpsInf*MaxCoo));
         end if;
      end loop;

      VV(VV'Last):=Norm(R,G)+Ball(Sup(EpsOne*MaxCo));
      return VV;
   end PProj;


   function PHProj(R: in Radius; D,N,II: in Integer; Eps: in Scalar; H: in Taylor; Name_Vec: in String; B: in Matrix) return Scalar is
      use Tay_IO;
      use Reps.IO;
      Dim: constant Integer := DD(N);
      G,GG: Taylor;
      HH: Taylor;
      V,Co: Vector(1..D);
      K: Integer;
      CCC,CC,MaxCo,Nrm1: Scalar;
      EV: Taylor;
   begin
      G:=H;
      GG:=H;
      K:=Dim;
      MaxCo:=Scal(Zero);
      for I in reverse -1..(N-1) loop
         for J in reverse IMax(0,I)..(N-1-I) loop

            if I=-1 or else I=J then
               HH:=Monom(R,I+1,J,ScalOne); --vector (xi+eta)
               Nrm1:=Norm(R,HH);
               CC:=Nrm1*Component(I+1,J,G);
               ZeroComponent(I+1,J,G);
            else
               HH:=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne); --vector (xi+eta)
               Nrm1:=Norm(R,HH);
               CCC:=Component(I+1,J,G);
               CC:=Nrm1*CCC;
               ZeroComponent(I+1,J,G);
               G:=G-CCC*Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne);
            end if;
	    
            if Sup(abs(CC))>Sup(MaxCo) then
               MaxCo:=abs(CC);
            end if;

            if K<=D then  Co(K):=CC;
            end if;

            K:=K-1;
         end loop;
      end loop;

      V:=B*Co;

      for I in 1..II loop
         SetZero(R,EV);
         Read(Name_Vec & Ind(I),EV);
         SetRho(R,EV);
         GG:=GG - ( V(I) + Eps*MaxCo  ) * EV;
      end loop;

      New_Line;
      Put("Rho(GG):");Txt_Put(Current_Output,Rho(GG));
      New_Line;
      Put("GG:");
      New_Line;
      Show(GG);

      return Norm(  R ,  GG ) ;
   end PHProj;


   function DDI(N1,N2: in Integer) return Integer is
   begin
      if N2>N1 then
         return N1;
      else
         return N2;
      end if;
   end DDI;


   procedure FindSS(Eps: in Radius; N: in Integer; Name_Vec,EV_Name,EVI_Name: in String; Rd,Rdz,Rdp1,Rp1,Rp11: in Radius; Ce,Delta_1, Delta_2: in Scalar; S,Z: in out Taylor  ) is
      use Ada.Text_IO;
      use Reps.IO;
      use Tay_IO;

      Dim: constant Integer := DD(N);
      Crit: Radius;
      EE,Lambda,Mu,T: Scalar;
      PV: Vector(1..Dim+1);
      R: constant Radius:=Rho(S);
      RS,H,H1,H2,HH,HO,S0: Taylor;
      F1,F2: File_Type;
      AI: Matrix(1..Dim,1..Dim);
      EV2: Vector(1..2*Dim);
   begin
      S0:=S;

      SetZero(R,H1);
      Read(Name_Vec & "001",H1);
      SetRho(R,H1);
      H1:=Inv(Norm(R,H1))*H1;


      SetZero(R,H2);
      Read(Name_Vec & "002",H2);
      SetRho(R,H2);
      H2:=Inv(Norm(R,H2))*H2;


      SetZero(AI);

      for II in AI'Range(2) loop
	 Open(F1,In_File,EV_Name & Ind(II));
	 Open(F2,In_File,EVI_Name & Ind(II));
	 EV2(II):=ScalZero;
	 for I in AI'Range(1) loop
	    Get(F1,EE);
	    EV2(II):=EV2(II)+abs(EE);

	    Get(F2,EE);
	    AI(I,II):=EE;
	 end loop;
	 Close(F2);
	 Close(F1);
      end loop;

      for II in AI'Range(2) loop
	 for I in AI'Range(1) loop
	    AI(I,II):=AI(I,II)*EV2(I);
	 end loop;
      end loop;

      SetZero(R,H);
      Crit:=One;
      while Crit > Eps loop
         HO:=H;
         PV:=PProj(R,AI'Last(1),N,Scal(Zero),Scal(Zero),H,AI);
         HH:=H-(PV(1)*H1+PV(2)*H2 );
         S:=S0+( Inv(Scal(One)-Delta_1)*PV(1)*H1+Inv(Scal(One)-Delta_2)*PV(2)*H2 )+HH ;

         RS:=S;
         RG(Eps,Rd,Rdz,Ce,Lambda, Mu,T,RS,Z);
         H:=H+RS-S;

         Crit:=MaxAbs(Norm(R,H-HO));
         New_Line;
         Put("Convergence in FindSS=");Txt_Put(Current_Output,Crit);
      end loop;


   end FindSS;







   procedure Proof(N: in Integer; Name_Vec,EV_Name,EVI_Name,Name_Bound,Name_Matrix,Name_New_Eps,Name_Eq_Const: in String;
				     Eps,Rd,Rdz,Rdp1,Rp1,Rp11: in Radius; Ce,Lambda,Mu,Delt: in Scalar; S,Z,H: in  Taylor; Compute: in Boolean:=True; Terms: in Integer) is
      use Ada.Text_IO;
      use Reps.IO;
      use Tay_IO;

      IL: Integer;
      NN: constant Integer:=PDeg-1;
      NM:  constant Integer:=3;
      NS: constant Integer:=3;
      Dim: constant Integer := DD(N);
      DimI: constant Integer:=Dim;
      Size: constant Integer:=DD(N);
      Iter: constant Integer := 50;
      K,Coun,Count: Integer;
      DB,New_Eps,Eq_Const,A_Norm, Bm,Cm,Delta1,BB_Norm,E1,E2,E3,BB,DerBC,RR1,RR2, M,EEE,EpsOne,EpsInf, Normm,EE,MaxOD,Cond,Nor: Scalar;
      B_Norm_One,M_Norm_One: Rep;
      B_Norm_Inf,M_Norm_Inf: Rep;
      Br,Bi: Vector(1..Dim);
      PV: Vector(1..Dim+1);
      OD: Matrix(1..Dim,1..Dim);

      Norr: Matrix(1..Size+1,1..Size+1);
      NDerMat,DerMat: Matrix(1..NM,1..NM);
      DerMat2: Matrix(1..2,1..2);

      HH,H1: Taylor;
      Num: String:="000";
      F1,F2,F3,F4,F5,F6: File_Type;
      Name_Matrix_1:  String:=Name_Matrix & "_1";
      Name_Matrix_2:  String:=Name_Matrix & "_2";
      Name_Matrix_3:  String:=Name_Matrix & "_3";

      Name_Matrix_11:  String:=Name_Matrix & "_11";
      Name_Matrix_22:  String:=Name_Matrix & "_22";
      Name_Matrix_33:  String:=Name_Matrix & "_33";

      Fact: constant Rep:=Rep(1.0E-8);


      ODD,AI,ADD,AD,ADK,ADDD,AII: Matrix(1..DimI,1..DimI);
      OD2: Matrix(1..2*DimI,1..2*DimI);
      EV2: Vector(1..2*DimI);
      Ev,Workk,EVV,EVVV: Vector(1..DimI);
      Pvtt: Pivot(1..DimI);
      R: constant Radius:=Rho(S);
   begin
      SetZero(R,HH);
      
      if  not Trunc then
	 if Terms=4 then
	    ---------------------------------------------------
	    Open(F1,In_File,Name_Matrix_1);
	    Open(F2,In_File,Name_Matrix_2);
	    Open(F3,In_File,Name_Matrix_3);
	    for J1 in 1..NM loop
	       for I1 in 1..NM loop
		  Get(F1,E1,False);
		  Get(F2,E2,False);
		  Get(F3,E3,False);
		  DerMat(I1,J1):=E1+E2+E3;
	       end loop;
	    end loop;
	    Close(F3);
	    Close(F2);
	    Close(F1);
	    -------------------------------------------------------

	    NDerMat:=DerMat;
	    NDerMat(1,1):=ResetCenter(NDerMat(1,1));
	    NDerMat(2,2):=ResetCenter(NDerMat(2,2));

	    DB:=Scal(Zero);
	    for I1 in 1..NM loop
	       A_Norm:=Scal(Zero);
	       for J1 in 1..NM loop
		  A_Norm:=abs(NDerMat(J1,I1))+A_Norm;
	       end loop;
	       if MaxAbs(A_Norm)>MaxAbs(DB) then
		  DB:=A_Norm;
	       end if;
	    end loop;
	    New_Line;
	    Put("Bound on the norm of the derivative in the Newton map=");Show(DB);

	    for II in AD'Range(2) loop
	       Open(F1,In_File,EV_Name & Ind(II));
	       EV2(II):=ScalZero;
	       for I in AD'Range(1) loop
		  Get(F1,EE);
		  AD(I,II):=EE;
		  EV2(II):=EV2(II)+abs(EE);
	       end loop;
	       Close(F1);
	    end loop;

	    Open(F1,In_File,Name_Eq_Const);
	    Get(F1,Eq_Const,False);
	    Close(F1);

	    Put("Equivalence constant (see proof of the Main Theorem for details)=");Show(Eq_Const);

	    Open(F1,In_File,Name_New_Eps);
	    Get(F1,New_Eps,False);
	    Close(F1);
	    
	    New_Line;
	    Put("Verification of the CMP (has to be a positive number less than one):");
	    Txt_Put(Current_Output,MaxAbs(Eq_Const*New_Eps/(Delt*(Scal(One)-DB) ) ) );
	    New_Line;

	    New_Line;
	    Put("LINEARIZED DERIVATIVE MATRIX:");
	    New_Line;
	    for J1 in 1..NM loop
	       for I1 in 1..NM loop
		  Show(DerMat(I1,J1));
	       end loop;
	       New_Line;
	    end loop;

	    DerMat2(2,2):=DerMat(NM,NM);
	    DerMat2(1,2):=Ball(MaxAbs(DerMat(1,NM))+MaxAbs(DerMat(2,NM)));
	    if MaxAbs(DerMat(3,1))>MaxAbs(DerMat(3,2)) then
	       DerMat2(2,1):=Ball(MaxAbs(DerMat(3,1)));
	    else
	       DerMat2(2,1):=Ball(MaxAbs(DerMat(3,2)));
	    end if;

	    DerMat2(1,1):=Ball(MaxAbs(DerMat(1,1))+MaxAbs(DerMat(2,1)));
	    BB:=Ball(MaxAbs(DerMat(1,2))+MaxAbs(DerMat(2,2)));
	    if MaxAbs(BB)>MaxAbs(DerMat2(1,1)) then
	       DerMat2(1,1):=BB;
	    end if;

	    DerBC:=(DerMat2(2,2)+DerMat2(1,2)*DerMat2(2,1)*(Scal(Two)+DerMat2(1,1)))/(Scal(One)+DerMat2(1,2)*DerMat2(2,1));


	    DerMat2(1,1):=DerMat(1,1);
	    DerMat2(2,1):=Ball(MaxAbs(DerMat(2,1))+MaxAbs(DerMat(3,1)));
	    if MaxAbs(DerMat(1,2))>MaxAbs(DerMat(1,3)) then
	       DerMat2(1,2):=Ball(MaxAbs(DerMat(1,2)));
	    else
	       DerMat2(1,2):=Ball(MaxAbs(DerMat(1,3)));
	    end if;

	    DerMat2(2,2):=Ball(MaxAbs(DerMat(2,2))+MaxAbs(DerMat(3,2)));
	    BB:=Ball(MaxAbs(DerMat(2,3))+MaxAbs(DerMat(3,3)));
	    if MaxAbs(BB)>MaxAbs(DerMat2(2,2)) then
	       DerMat2(2,2):=BB;
	    end if;

	    Bm:=Two*Ball(MaxAbs(DerMat2(1,2)))/(DerMat2(1,1)-DerMat2(2,2));
	    Cm:=Two*Ball(MaxAbs(DerMat2(2,1)))/(DerMat2(1,1)-DerMat2(2,2));

	    Delta1:=DerMat2(1,1)+(Bm*DerMat2(2,1)-Cm*DerMat2(1,2)+Cm*Bm*(DerMat2(2,2)-DerMat2(1,1)) )/(Bm*Cm-Scal(One));

	    New_Line;
	    Put("BOUND ON THE SPECTRAL RADIUS OF THE DERIVATIVE IN THE CONTRACTING SUBSPACE:");
	    Txt_Put(Current_Output,MaxAbs(DerBC));
	    New_Line;
	    New_Line;
	    Put("BOUND ON THE LARGEST EIGENVALUE: (");
	    Txt_Put(Current_Output,Inf(Delta1) );
	    Put(",");
	    Txt_Put(Current_Output,Sup(Delta1) );
	    Put(")");

	    Create(F1,Out_File,Name_Bound);
	    Put(F1,DerBC,False);
	    Close(F1);

	    return;
	 end if;


         SetZero(AI);
         SetZero(AD);

         for II in AD'Range(2) loop
            Open(F1,In_File,EV_Name & Ind(II));
            EV2(II):=ScalZero;
            for I in AD'Range(1) loop
               Get(F1,EE);
               AD(I,II):=EE;
               EV2(II):=EV2(II)+abs(EE);
            end loop;
            Close(F1);
         end loop;




         B_Norm_One:=Zero;
         for II in AI'Range(2) loop
            Open(F2,In_File,EVI_Name & Ind(II));
            BB_Norm:=Scal(Zero);
            for I in AI'Range(1) loop
               Get(F2,EEE);
               AI(I,II):=EEE;
               BB_Norm:=BB_Norm+abs(EEE);
            end loop;
            if MaxAbs(BB_Norm)>B_Norm_One then B_Norm_One:=MaxAbs(BB_Norm); end if;
            Close(F2);
         end loop;
         New_Line;
         Put("||B_0||_1=");Txt_Put(Current_Output,B_Norm_One);

         B_Norm_Inf:=Zero;
         for I in AI'Range(1) loop
            BB_Norm:=Scal(Zero);
            for II in AI'Range(2) loop
               BB_Norm:=BB_Norm+abs(AI(I,II));
            end loop;
            if MaxAbs(BB_Norm)>B_Norm_Inf then B_Norm_Inf:=MaxAbs(BB_Norm); end if;
         end loop;
         New_Line;
         Put("||B_0||_infty=");Txt_Put(Current_Output,B_Norm_Inf);

         ADD:=AD*AI;


         M_Norm_One:=Zero;
         for I in ADD'Range(1) loop
            Normm:=Scal(Zero);
            for J in ADD'Range(2) loop
               if I=J then
                  ADD(J,I):=ADD(J,I)-Scal(One);
               end if;
               Normm:=Normm+abs(ADD(J,I));
            end loop;
            if Sup(Normm)>M_Norm_One then
               M_Norm_One:=Sup(Normm);
            end if;
         end loop;


         M_Norm_Inf:=Zero;
         for I in ADD'Range(1) loop
            Normm:=Scal(Zero);
            for J in ADD'Range(2) loop
               Normm:=Normm+abs(ADD(I,J));
            end loop;
            if Sup(Normm)>M_Norm_Inf then
               M_Norm_Inf:=Sup(Normm);
            end if;
         end loop;
	 EpsInf:=B_Norm_Inf*Ball(Sup(Scal(M_Norm_Inf)/(Scal(One)-Scal(M_Norm_Inf))));
         EpsOne:=B_Norm_One*Ball(Sup(Scal(M_Norm_One)/(Scal(One)-Scal(M_Norm_One))));
         New_Line;
         Put("Inverting the matrix: l_1 bound in CMP=");
         Txt_Put(Current_Output,Sup(EpsOne));
         New_Line;
         Put("Inverting the matrix: l_infty bound in CMP=");
         Txt_Put(Current_Output,Sup(EpsInf));



         for II in AI'Range(2) loop
            for I in AI'Range(1) loop
               AI(I,II):=AI(I,II)*EV2(I);
            end loop;
         end loop;


         Eq_Const:=Scal(Zero);
         for I in AD'Range(1) loop
            A_Norm:=Scal(Zero);
            for II in AD'Range(2) loop
               EE:=AD(I,II);
               if abs(EE)>A_Norm then
                  A_Norm:=abs(EE);
               end if;
            end loop;
            Eq_Const:=Eq_Const+A_Norm;
         end loop;

         PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,H,AI);
         New_Eps:=Scal(Zero);
         for I in 1..Dim loop
            New_Eps:=New_Eps+abs(PV(I));
         end loop;
         New_Eps:=New_Eps+abs(PV(PV'Last));
         New_Line;
         Put("New norm of the difference="); Txt_Put(Current_Output,MaxAbs(New_Eps));
         New_Line;
         Put("Equivalence constant="); Txt_Put(Current_Output,MaxAbs(Eq_Const));

         Create(F1,Out_File,Name_Eq_Const);
         Put(F1,Eq_Const,False);
         Close(F1);

         Create(F1,Out_File,Name_New_Eps);
         Put(F1,New_Eps,False);
         Close(F1);

         if Terms=0 then
            SetZero(Norr);
            ---------------------------------------------------------
            New_Line;
            Put("DOING DERIVATIVE BOUNDS ON LOWER-ORDER TERMS...");
	    New_Line;
	    ----------------------------------------------------------
	    
            for II in AI'Range(1) loop

               SetZero(R,HH);
	       New_Line;
	       Read(Name_Vec & Ind(II),HH);
               SetRho(R,HH);
               HH:=Inv(Norm(R,HH))*HH;
	       Put("Doing DRGh...");
               DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,HH);
	       Put("Done");
	       
               PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,HH,AI);
               for JJ in 1..Size loop
                  Norr(JJ,II):=PV(JJ);
               end loop;
               Norr(Size+1,II):=PV(Size+1);

            end loop;
            Put("DONE");

            ---------------------------------------------

            for I1 in 1..NM-1 loop
               for J1 in 1..NM-1 loop
                  DerMat(I1,J1):=Norr(I1,J1);
               end loop;
            end loop;


            --Last row
            for I1 in 1..NM-1 loop
               DerMat(NM,I1):=Scal(Zero);
               for J1 in NM..Size+1 loop
                  DerMat(NM,I1):=DerMat(NM,I1)+abs(Norm(Norr(J1,I1)));
               end loop;
            end loop;


            --Last column
            for J1 in 1..NM-1 loop
	       M:=Scal(Zero);
               for I1 in NM..Size+1 loop
		  if MaxAbs(Norm(Norr(J1,I1)))>MaxAbs(M) then
                     M:=Norm(Norr(J1,I1));
                  end if;
               end loop;
               DerMat(J1,NM):=M;
            end loop;

            DerMat(NM,NM):=Scal(Zero);
            for I1 in NM..Size+1 loop
               M:=Scal(Zero);
               for J1 in NM..Size+1 loop
                  M:=M+Norm(Norr(J1,I1));
               end loop;
               if MaxAbs(Norm(M))>MaxAbs(Norm(DerMat(NM,NM))) then
                  DerMat(NM,NM):=M;
               end if;
            end loop;

            New_Line;
            Put("LINEARIZED DERIVATIVE MATRIX:");
            New_Line;
            for J1 in 1..NM loop
               for I1 in 1..NM loop
                  Show(DerMat(I1,J1));
               end loop;
               New_Line;
            end loop;
            -----------------------------------------------

            ----------------------------------------------------------
            New_Line;
            Put("DOING DERIVATIVE BOUNDS ON HIGHER-ORDER TERMS...");
	    New_Line;
	    ----------------------------------------------------------

            RR1:=Scal(R)**(NN+1);
            for I in -1..NN loop
               New_Line;
               if I=-1 or else I=(NN-I) then
                  H1:=Inv(RR1)*Monom(R,I+1,NN-I,Ball(Zero,One));
               else
                  RR2:=Scal(NN+2)/Scal(NN-I+1)*RR1;
                  H1:=Inv(RR2)*(Monom(R,I+1,NN-I,Ball(Zero,One))+Scal(I+1)/Scal(NN-I+1)*Monom(R,NN-I+1,I,Ball(Zero,One)));
               end if;

               H1:=Scal(Fact)*H1;
	       New_Line;
	       Put("Doing DRGh...");
               DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,H1);
	       Put("Done");
	       H1:=Inv(Scal(Fact))*H1;

               PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,H1,AI);

               for JJ in 1..Size loop
                  if MaxAbs(PV(JJ))>MaxAbs(Norr(JJ,Size+1)) then
                     Norr(JJ,Size+1):=PV(JJ);
                  end if;
               end loop;

               M:=PV(Size+1);
               if MaxAbs(M)>MaxAbs(Norr(Size+1,Size+1)) then
                  Norr(Size+1,Size+1):=M;
               end if;
            end loop;
            Put("DONE");
	    
            ---------------------------------------------

            for I1 in 1..NM-1 loop
               for J1 in 1..NM-1 loop
                  DerMat(I1,J1):=Norr(I1,J1);
               end loop;
            end loop;


            --Last row
            for I1 in 1..NM-1 loop
               DerMat(NM,I1):=Scal(Zero);
               for J1 in NM..Size+1 loop
                  DerMat(NM,I1):=DerMat(NM,I1)+abs(Norm(Norr(J1,I1)));
               end loop;
            end loop;


            --Last column
            for J1 in 1..NM-1 loop
	       M:=Scal(Zero);
               for I1 in NM..Size+1 loop
		  if MaxAbs(Norm(Norr(J1,I1)))>MaxAbs(M) then
                     M:=Norm(Norr(J1,I1));
                  end if;
               end loop;
               DerMat(J1,NM):=M;
            end loop;


            DerMat(NM,NM):=Scal(Zero);
            for I1 in NM..Size+1 loop
               M:=Scal(Zero);
               for J1 in NM..Size+1 loop
                  M:=M+Norm(Norr(J1,I1));
               end loop;
               if MaxAbs(Norm(M))>MaxAbs(Norm(DerMat(NM,NM))) then
                  DerMat(NM,NM):=M;
               end if;
            end loop;

            New_Line;
            Put("LINEARIZED DERIVATIVE MATRIX:");
            New_Line;
            for J1 in 1..NM loop
               for I1 in 1..NM loop
                  Show(DerMat(I1,J1));
               end loop;
               New_Line;
            end loop;
            -----------------------------------------------

            -------------------------------------------------------
            New_Line;
            Put("DOING DERIVATIVE BOUND ON NON-EIGENVECTOR TERMS...");
	    New_Line;
            -------------------------------------------------------

            for I in  -1..(NN-1) loop
               for J in IMax(0,I)..(NN-1-I) loop
                  if (I+J)>(N-1) then
                     if I=-1 or else I=J then
                        H1:=Monom(R,I+1,J,Scal(One));
                     else
                        H1:=(Monom(R,I+1,J,Scal(One))+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,Scal(One)));
                     end if;


                     H1:=Inv(Norm(R,H1))*H1;
                     New_Line;
		     Put("Doing DRGh...");
		     DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,H1);
		     Put("Done");
		     PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,H1,AI);
                     for JJ in 1..Size loop
                        if MaxAbs(PV(JJ))>MaxAbs(Norr(JJ,Size+1)) then
                           Norr(JJ,Size+1):=PV(JJ);
                        end if;
                     end loop;

                     M:=PV(Size+1);
                     if MaxAbs(M)>MaxAbs(Norr(Size+1,Size+1)) then
                        Norr(Size+1,Size+1):=M;
                     end if;

                  end if;

               end loop;
            end loop;
            Put("DONE");

            ---------------------------------------------

            for I1 in 1..NM-1 loop
               for J1 in 1..NM-1 loop
                  DerMat(I1,J1):=Norr(I1,J1);
               end loop;
            end loop;


            --Last row
            for I1 in 1..NM-1 loop
               DerMat(NM,I1):=Scal(Zero);
               for J1 in NM..Size+1 loop
                  DerMat(NM,I1):=DerMat(NM,I1)+abs(Norm(Norr(J1,I1)));
               end loop;
            end loop;


            --Last column
            for J1 in 1..NM-1 loop
               M:=Scal(Zero);
               for I1 in NM..Size+1 loop
		  if MaxAbs(Norm(Norr(J1,I1)))>MaxAbs(M) then
                     M:=Norm(Norr(J1,I1));
                  end if;
               end loop;
               DerMat(J1,NM):=M;
            end loop;


            DerMat(NM,NM):=Scal(Zero);
            for I1 in NM..Size+1 loop
               M:=Scal(Zero);
               for J1 in NM..Size+1 loop
                  M:=M+Norm(Norr(J1,I1));
               end loop;
               if MaxAbs(Norm(M))>MaxAbs(Norm(DerMat(NM,NM))) then
                  DerMat(NM,NM):=M;
               end if;
            end loop;
            ---------------------------------------------------
            New_Line;
            Put("LINEARIZED DERIVATIVE MATRIX:");
            New_Line;
            Create(F1,Out_File,Name_Matrix);
            for J1 in 1..NM loop
               for I1 in 1..NM loop
                  Show(DerMat(I1,J1));
                  Put(F1,DerMat(I1,J1),False);
                  New_Line(F1);
               end loop;
               New_Line;
            end loop;
            Close(F1);
            -------------------------------------------------------
	    
            DerMat2(2,2):=DerMat(NM,NM);
            DerMat2(1,2):=Ball(MaxAbs(DerMat(1,NM))+MaxAbs(DerMat(2,NM)));
            if MaxAbs(DerMat(3,1))>MaxAbs(DerMat(3,2)) then
               DerMat2(2,1):=Ball(MaxAbs(DerMat(3,1)));
            else
               DerMat2(2,1):=Ball(MaxAbs(DerMat(3,2)));
            end if;

            DerMat2(1,1):=Ball(MaxAbs(DerMat(1,1))+MaxAbs(DerMat(2,1)));
            BB:=Ball(MaxAbs(DerMat(1,2))+MaxAbs(DerMat(2,2)));
            if MaxAbs(BB)>MaxAbs(DerMat2(1,1)) then
               DerMat2(1,1):=BB;
            end if;
            DerBC:=(DerMat2(2,2)+DerMat2(1,2)*DerMat2(2,1)*(Scal(Two)+DerMat2(1,1)))/(Scal(One)+DerMat2(1,2)*DerMat2(2,1));

            New_Line;
            Put("DER BOUND IN THE CONTRACTING SUBSPACE:");
            Show(DerBC);

            Create(F1,Out_File,Name_Bound);
            Put(F1,DerBC,False);
            Close(F1);

         elsif Terms=1 then
	    
            SetZero(Norr);

            ---------------------------------------------------------
            New_Line;
            Put("DOING DERIVATIVE BOUNDS ON LOWER-ORDER TERMS...");
	    New_Line;
            ----------------------------------------------------------

            for II in AI'Range(1) loop

               SetZero(R,HH);
	       New_Line;
	       Read(Name_Vec & Ind(II),HH);
               SetRho(R,HH);
               HH:=Inv(Norm(R,HH))*HH;
	       Put("Doing DRGh...");
               DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,HH);
	       Put("Done");
               New_Line;
               Put("Eigenvector "); Txt_Put(Current_Output,Rep(II));Put(" out of "); Txt_Put(Current_Output,Rep(AI'Last(1)) ); Put(" done");
               if II <=Size then
                  PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,HH,AI);
                  for JJ in 1..Size loop
                     Norr(JJ,II):=PV(JJ);
                  end loop;
                  Norr(Size+1,II):=PV(Size+1);
               elsif II>Size then
                  PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,HH,AI);
                  for JJ in 1..Size loop
                     Nor:=PV(JJ);
                     if MaxAbs(Nor)>MaxAbs(Norr(JJ,Size+1)) then
                        Norr(JJ,Size+1):=Nor;
                     end if;
                  end loop;
                  Nor:=PV(Size+1);
                  if MaxAbs(Nor)>MaxAbs(Norr(Size+1,Size+1)) then
                     Norr(Size+1,Size+1):=Nor;
                  end if;
               end if;
            end loop;
            Put("DONE");
            ---------------------------------------------

            for I1 in 1..NM-1 loop
               for J1 in 1..NM-1 loop
                  DerMat(I1,J1):=Norr(I1,J1);
               end loop;
            end loop;

            --Last row
            for I1 in 1..NM-1 loop
               DerMat(NM,I1):=Scal(Zero);
               for J1 in NM..Size+1 loop
                  DerMat(NM,I1):=DerMat(NM,I1)+abs(Norm(Norr(J1,I1)));
               end loop;
            end loop;

            --Last column
            for J1 in 1..NM-1 loop
               M:=Scal(Zero);
               for I1 in NM..Size+1 loop
                  if MaxAbs(Norm(Norr(J1,I1)))>MaxAbs(M) then
                     M:=Norm(Norr(J1,I1));
                  end if;
               end loop;
               DerMat(J1,NM):=M;
            end loop;
            DerMat(NM,NM):=Scal(Zero);
            for I1 in NM..Size+1 loop
               M:=Scal(Zero);
               for J1 in NM..Size+1 loop
                  M:=M+Norm(Norr(J1,I1));
               end loop;
               if MaxAbs(Norm(M))>MaxAbs(Norm(DerMat(NM,NM))) then
                  DerMat(NM,NM):=M;
               end if;
            end loop;
            ---------------------------------------------------
            New_Line;
            Put("LINEARIZED DERIVATIVE MATRIX:");
            New_Line;
            Create(F1,Out_File,Name_Matrix_1);
            for J1 in 1..NM loop
               for I1 in 1..NM loop
                  Show(DerMat(I1,J1));
                  Put(F1,DerMat(I1,J1),False);
                  New_Line(F1);
               end loop;
               New_Line;
            end loop;
            Close(F1);
            -------------------------------------------------------

         elsif Terms=2 then

            SetZero(Norr);

            ----------------------------------------------------------
            New_Line;
            Put("DOING DERIVATIVE BOUNDS ON HIGHER-ORDER TERMS...");
	    New_Line;
            ----------------------------------------------------------

            RR1:=Scal(R)**(NN+1);
            for I in -1..NN loop

               if I=-1 or else I=(NN-I) then
                  H1:=Inv(RR1)*Monom(R,I+1,NN-I,Ball(Zero,One));
               else
                  RR2:=Scal(NN+2)/Scal(NN-I+1)*RR1;
                  H1:=Inv(RR2)*(Monom(R,I+1,NN-I,Ball(Zero,One))+Scal(I+1)/Scal(NN-I+1)*Monom(R,NN-I+1,I,Ball(Zero,One)));
               end if;

               H1:=Scal(Fact)*H1;
	       New_Line;
	       Put("Doing DRGh...");
	       DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,H1);
	       Put("Done");
	       H1:=Inv(Scal(Fact))*H1;

               New_Line;
               Put("Monomial "); Txt_Put(Current_Output,Rep(I));Put(" out of "); Txt_Put(Current_Output,Rep(NN));Put(" done");
               PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,H1,AI);
               for JJ in 1..Size loop
                  if MaxAbs(PV(JJ))>MaxAbs(Norr(JJ,Size+1)) then
                     Norr(JJ,Size+1):=PV(JJ);
                  end if;
               end loop;
               M:=PV(Size+1);
               if MaxAbs(M)>MaxAbs(Norr(Size+1,Size+1)) then
                  Norr(Size+1,Size+1):=M;
               end if;
            end loop;
            Put("DONE");

            ---------------------------------------------

            for I1 in 1..NM-1 loop
               for J1 in 1..NM-1 loop
                  DerMat(I1,J1):=Norr(I1,J1);
               end loop;
            end loop;

            --Last row
            for I1 in 1..NM-1 loop
               DerMat(NM,I1):=Scal(Zero);
               for J1 in NM..Size+1 loop
                  DerMat(NM,I1):=DerMat(NM,I1)+abs(Norm(Norr(J1,I1)));
               end loop;
            end loop;

            --Last column
            for J1 in 1..NM-1 loop
               M:=Scal(Zero);
               for I1 in NM..Size+1 loop
		  if MaxAbs(Norm(Norr(J1,I1)))>MaxAbs(M) then
                     M:=Norm(Norr(J1,I1));
                  end if;
               end loop;
               DerMat(J1,NM):=M;
            end loop;
	    
            DerMat(NM,NM):=Scal(Zero);
            for I1 in NM..Size+1 loop
               M:=Scal(Zero);
               for J1 in NM..Size+1 loop
                  M:=M+Norm(Norr(J1,I1));
               end loop;
               if MaxAbs(Norm(M))>MaxAbs(Norm(DerMat(NM,NM))) then
                  DerMat(NM,NM):=M;
               end if;
            end loop;
	    
            ---------------------------------------------------
            New_Line;
            Put("LINEARIZED DERIVATIVE MATRIX:");
            New_Line;
            Create(F1,Out_File,Name_Matrix_2);
            for J1 in 1..NM loop
               for I1 in 1..NM loop
                  Show(DerMat(I1,J1));
                  Put(F1,DerMat(I1,J1),False);
                  New_Line(F1);
               end loop;
               New_Line;
            end loop;
            Close(F1);
            -------------------------------------------------------

         elsif Terms=3 then

            SetZero(Norr);

            -------------------------------------------------------
            New_Line;
            Put("DOING DERIVATIVE BOUND ON NON-EIGENVECTOR TERMS...");
	    New_Line;
            -------------------------------------------------------

            Count:=0;
            for I in  -1..(NN-1) loop
               for J in IMax(0,I)..(NN-1-I) loop
                  if (I+J)>(N-1) then
                     Count:=Count+1;
                  end if;
               end loop;
            end loop;
	    
            Coun:=0;
            for I in  -1..(NN-1) loop
               for J in IMax(0,I)..(NN-1-I) loop
                  if (I+J)>(N-1) then
                     Coun:=Coun+1;
                     if I=-1 or else I=J then
                        H1:=Monom(R,I+1,J,Scal(One));
                     else
                        H1:=(Monom(R,I+1,J,Scal(One))+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,Scal(One)));
                     end if;
                     H1:=Inv(Norm(R,H1))*H1;
		     New_Line;
		     Put("Doing DRGh...");
		     DRGh(Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,H1);
		     Put("Done");
		     New_Line;
                     Put("Term "); Txt_Put(Current_Output,Rep(Coun));Put(" out of "); Txt_Put(Current_Output,Rep(Count));Put(" done");

                     PV:=PProj(R,AI'Last(1),N,EpsOne,EpsInf,H1,AI);
                     for JJ in 1..Size loop
                        if MaxAbs(PV(JJ))>MaxAbs(Norr(JJ,Size+1)) then
                           Norr(JJ,Size+1):=PV(JJ);
                        end if;
                     end loop;
                     M:=PV(Size+1);
                     if MaxAbs(M)>MaxAbs(Norr(Size+1,Size+1)) then
                        Norr(Size+1,Size+1):=M;
                     end if;
                  end if;
               end loop;
            end loop;
            Put("DONE");

            ---------------------------------------------

            for I1 in 1..NM-1 loop
               for J1 in 1..NM-1 loop
                  DerMat(I1,J1):=Norr(I1,J1);
               end loop;
            end loop;

            --Last row
            for I1 in 1..NM-1 loop
               DerMat(NM,I1):=Scal(Zero);
               for J1 in NM..Size+1 loop
                  DerMat(NM,I1):=DerMat(NM,I1)+abs(Norm(Norr(J1,I1)));
               end loop;
            end loop;

            --Last column
            for J1 in 1..NM-1 loop
               M:=Scal(Zero);
               for I1 in NM..Size+1 loop
		  if MaxAbs(Norm(Norr(J1,I1)))>MaxAbs(M) then
                     M:=Norm(Norr(J1,I1));
                  end if;
               end loop;
               DerMat(J1,NM):=M;
            end loop;
	    DerMat(NM,NM):=Scal(Zero);
            for I1 in NM..Size+1 loop
               M:=Scal(Zero);
               for J1 in NM..Size+1 loop
                  M:=M+Norm(Norr(J1,I1));
               end loop;
               if MaxAbs(Norm(M))>MaxAbs(Norm(DerMat(NM,NM))) then
                  DerMat(NM,NM):=M;
               end if;
            end loop;
	    
            ---------------------------------------------------
            New_Line;
            Put("LINEARIZED DERIVATIVE MATRIX:");
            New_Line;
            Create(F1,Out_File,Name_Matrix_3);
            for J1 in 1..NM loop
               for I1 in 1..NM loop
                  Show(DerMat(I1,J1));
                  Put(F1,DerMat(I1,J1),False);
                  New_Line(F1);
               end loop;
               New_Line;
            end loop;
            Close(F1);
            -------------------------------------------------------
         end if;
      else
         if Compute then
            New_Line;
            Put("Doing eigenvalues...");
            DRGG(N,Eps,Rd,Rdz,Rdp1,Rp1,Rp11,Ce,Lambda,Mu,S,Z,OD);
            Eigenvalues(OD,Br,Bi);
            Put("done");

            for I in ODD'Range(1) loop
               for J in ODD'Range(2) loop
                  ODD(I,J):=OD(I,J);
               end loop;
            end loop;

            IL:=1;
            while IL<= AI'Last(1) loop

               if Bi(IL)=Scal(Zero) then
                  New_Line;
                  Put("Doing eigenvector " & Ind(IL) & " ...");
                  Eigenvector(ODD,Center(Br(IL)),EV, Scal(1.0E-20));
                  Create(F1,Out_File,EV_Name & Ind(IL));
                  for I in AD'Range(1) loop
                     Put(F1,EV(I));
                     New_Line(F1);
                     AD(I,IL):=EV(I);
                  end loop;
                  Close(F1);
                  Put("done");

                  K:=1;
                  SetZero(R,HH);
                  for I in -1..(N-1) loop
                     for J in IMax(0,I)..(N-1-I) loop
                        if K<=EV'Last then
                           if I=-1 or else I=J then
                              H1:=Monom(R,I+1,J,ScalOne);
                           else
                              H1:=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne);
                           end if;
                           H1:=Inv(Norm(R,H1))*H1;
                           HH:=HH+EV(K)*H1;
                        end if;
                        K:=K+1;
                     end loop;
                  end loop;

                  New_Line;
                  HH:=Inv(Norm(R,HH))*HH;
                  Write(Name_Vec &  Ind(IL),HH);
                  IL:=IL+1;
               else
                  for I in ODD'Range(1) loop
                     for J in ODD'Range(2) loop
                        OD2(I,J):=ODD(I,J);
                        OD2(I+ODD'Last(1),J+ODD'Last(2)):=ODD(I,J);
                        if I=J then
                           OD2(I,J+ODD'Last(2)):=Bi(IL);
                           OD2(I+ODD'Last(1),J):=-Bi(IL);
                        end if;
                     end loop;
                  end loop;
                  New_Line;
                  Put("Doing eigenvectors " & Ind(IL) & " and " & Ind(IL+1) & " ...");
                  Eigenvector(OD2,Center(Br(IL)),EV2, Scal(1.0E-20));

                  Create(F1,Out_File,EV_Name & Ind(IL));
                  for I in AD'Range(1) loop
                     Put(F1,EV2(I)); New_Line(F1);
                     AD(I,IL):=EV2(I);
                  end loop;
                  Close(F1);
                  Put("done");

                  K:=1;
                  SetZero(R,HH);
                  for I in -1..(N-1) loop
                     for J in IMax(0,I)..(N-1-I) loop
                        if K<=EV'Last then
                           if I=-1 or else I=J then
                              H1:=Monom(R,I+1,J,ScalOne);
                           else
                              H1:=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne);
                           end if;
                           H1:=Inv(Norm(R,H1))*H1;
                           HH:=HH+EV2(K)*H1;
                        end if;
                        K:=K+1;
                     end loop;
                  end loop;

                  New_Line;
                  HH:=Inv(Norm(R,HH))*HH;
                  Write(Name_Vec &  Ind(IL),HH);

                  IL:=IL+1;

                  Create(F1,Out_File,EV_Name & Ind(IL));
                  for I in AD'Range(1) loop
                     Put(F1,EV2(I+ODD'Last(1)));
                     New_Line(F1);
                     AD(I,IL):=EV2(I+ODD'Last(1));
                  end loop;
                  Close(F1);
                  Put("done");

                  K:=1;
                  SetZero(R,HH);
                  for I in -1..(N-1) loop
                     for J in IMax(0,I)..(N-1-I) loop
                        if K<=EV'Last then
                           if I=-1 or else I=J then
                              H1:=Monom(R,I+1,J,ScalOne);
                           else
                              H1:=Monom(R,I+1,J,ScalOne)+Scal(I+1)/Scal(J+1)*Monom(R,J+1,I,ScalOne);
                           end if;
                           H1:=Inv(Norm(R,H1))*H1;
                           HH:=HH+EV2(K+ODD'Last(1))*H1;
                        end if;
                        K:=K+1;
                     end loop;
                  end loop;

                  New_Line;
                  HH:=Inv(Norm(R,HH))*HH;
                  Write(Name_Vec &  Ind(IL),HH);

                  IL:=IL+1;

               end if;
            end loop;

            for I in AD'Range(1) loop
               for J in AD'Range(2) loop
                  ADDD(I,J):=AD(I,J);
               end loop;
            end loop;

            ADK:=ADDD;

            Decomp(ADDD,Cond,Pvtt,Workk);

            for I in ADDD'Range(1) loop
               SetZero(EVV);
               EVV(I):=Scal(One);
               EV:=EVV;
               Solve(ADDD,EVV,Pvtt);


               EVVV:=ADK*EVV-EV;
               MaxOD:=Scal(Zero);
               for J1 in EVV'Range loop
                  if MaxAbs(EVV(J1))>MaxAbs(MaxOD) then
                     MaxOd:=abs(EVV(J1));
                  end if;
               end loop;

               Create(F1,Out_File,EVI_Name & Ind(I));
               for J in AII'Range(2) loop
                  AII(J,I):=EVV(J);
                  Put(F1,EVV(J));
                  New_Line(F1);
               end loop;
               Close(F1);

            end loop;

            ADDD:=ADK*AII;
            MaxOD:=Scal(Zero);
            for I in ADDD'Range(1) loop
               for J in ADDD'Range(2) loop
                  if I=J  then
                     ADDD(I,J):=ADDD(I,J)-Scal(One);
                  end if;
                  if abs(ADDD(I,J))>MaxOD then
                     MaxOD:=abs(ADDD(I,J));
                  end if;
               end loop;
            end loop;

            New_Line;
            New_Line;
            Put("l_infty norm of the A*A^{-1}=");Txt_Put(Current_Output,Sup(MaxOD));

         else

            for II in 1..Dim loop
               Open(F1,In_File,EV_Name & Ind(II));
               for I in EV'Range loop
                  Get(F1,EE);
                  AD(I,II):=EE;
               end loop;
               Close(F1);
            end loop;

            for I in 1..DimI loop
               for J in 1..DimI loop
                  ADDD(I,J):=AD(I,J);
               end loop;
            end loop;

            ADK:=ADDD;
            Decomp(ADDD,Cond,Pvtt,Workk);

            for I in ADDD'Range(1) loop
               SetZero(EVV);
               EVV(I):=Scal(One);
               Solve(ADDD,EVV,Pvtt);

               Create(F1,Out_File,EVI_Name & Ind(I));
               for J in AII'Range(2) loop
                  AII(J,I):=EVV(J);
                  Put(F1,EVV(J)); New_Line(F1);
               end loop;
               Close(F1);
            end loop;


            ADDD:=ADK*AII;
            MaxOD:=Scal(Zero);
            for I in ADDD'Range(1) loop
               for J in ADDD'Range(2) loop
                  if I=J  then
                     ADDD(I,J):=ADDD(I,J)-Scal(One);
                  end if;
                  if abs(ADDD(I,J))>MaxOD then
                     MaxOD:=abs(ADDD(I,J));
                  end if;
               end loop;
            end loop;

            New_Line;
            New_Line;
            Put("l_infty norm of the A*A^{-1}=");Txt_Put(Current_Output,Sup(MaxOD));

         end if;
      end if;

   end Proof;
   
   
end Doubling;

