with Reps, Scalars, Scalars.IO, Vectors, Vectors.Num, Taylors, Taylors.IO,Ada.Text_IO;
use Reps,Ada.Text_IO;

generic

   PDeg: in Natural;
   CutOff: in Natural;
   --   Ce: in Scalar:=Scal(-5.97457559741896682E-003);
   with package Numbers is new Scalars (<>);
   with package Numbers_IO is new Numbers.IO (<>);
   
package Doubling is
   
   use Numbers, Numbers_IO;
   
   package Tay is new Taylors (PDeg => PDeg,CutOff=>CutOff, Numbers => Numbers);
   package Tay_IO is new Tay.IO (Numbers_IO => Numbers_IO);
   use Tay;
   
   function CompZ(F1,F2,Z: Taylor) return Taylor;                        -- F1(Z,.)+F2(.,Z)
   function DCompZ(Rd: Radius; F1,F2,Z: Taylor) return Taylor;            -- D1F1(Z,.)+D2F2(.,Z)

   --procedures for symmetric functions
   function CompZ(S,X,Y,Z: Taylor) return Taylor;                        -- S(X,Z)+S(Y,Z)
   function DCompZ(Rd: Radius; S,X,Y,Z: Taylor) return Taylor;            -- D2S(X,Z)+D2S(Y,Z)
   procedure  FindZ(Epsi: in out Radius; Rd: in Radius; S: in Taylor; Z: in out Taylor); -- solve S(X,Z)+S(Y,Z)=0
   procedure FindT(S,Z: in Taylor;  Ce: in Scalar; T: out Scalar);
   procedure FindLambda(Rdz: in Radius; Z,S0,P: in Taylor; Lambda: out Scalar); -- solve S(Lambda-A,1-A)+S(-A,1-A)=0
   procedure RG(Eps,Rd,Rdz: in Radius; Ce: in Scalar; Lambda,Mu,T: out Scalar; S,Z: in out Taylor  ); --S->R(S), normalization a-la Denis
   procedure PG(Eps,Rd: in Radius;  S,Z: in out Taylor;  Ce: in Scalar; T: out Scalar);
   procedure DLambdah(Rp1: in Radius; Lambda: in Scalar; P,DP: in Taylor; DL: out Scalar); -- DL:=Lambda'[S].H
   function DLambdah(Rp1: Radius; Lambda: Scalar; P,DP: Taylor) return Scalar;
   procedure DMuh(Rdp1,Rp1,Rp11: in Radius; Lambda,DL: in Scalar; P,DP: in  Taylor; DM: out Scalar); --DM:=Mu'[S].H
   function DMuh(Rdp1,Rp1,Rp11: Radius; Lambda,DL: Scalar; P,DP: Taylor) return Scalar;
   procedure DZh(Rd: in Radius; S,Z: in Taylor; H: in out Taylor); --H ->Z'[S].H
   procedure DTh(Rd: in Radius; Ce,T: in Scalar; S,Z,H,DZ: in Taylor;  DT: out Scalar);
   procedure DPGh(Rd,Rdz: in Radius; Ce,T: in Scalar; S: in Taylor; Z: in Taylor; H: in out Taylor);
   procedure Proof(N: in Integer; Name_Vec,EV_Name,EVI_Name,Name_Bound,Name_Matrix,Name_New_Eps,Name_Eq_Const: in String; Eps,Rd,Rdz,Rdp1,Rp1,Rp11: in Radius;
                                                                                                                             Ce,Lambda,Mu,Delt: in  Scalar; S,Z,H: in  Taylor; Compute: in Boolean:=True; Terms: in Integer);
   
   procedure FindS(Eps: in Radius; N: in Integer; Ce: in Scalar; Rd,Rdz,Rdp1,Rp1,Rp11: in Radius;  S,Z: in out Taylor  ); -- Finds S, s.t. P_N[R(S)-S]=0
   procedure FindSS(Eps: in Radius; N: in Integer; Name_Vec,EV_Name,EVI_Name: in String; Rd,Rdz,Rdp1,Rp1,Rp11: in Radius; Ce,Delta_1, Delta_2: in Scalar; S,Z: in out Taylor  ); --Newton Map for R


private
   ScalZero:  constant Scalar := Scal(Zero);
   ScalOne:  constant Scalar := Scal(One);
   ScalTwo:  constant Scalar := Scal(Two);
   ScalHalf:  constant Scalar := Scal(Half);
   ScalPi:  constant Scalar := Scal(3.14159265358979323846,3.14159265358979323847);
   Trunc: constant Boolean := IsNumeric(ScalZero);
   
   package Vec is new Vectors (Scalar => Scalar, Zero_Scalar => ScalZero);
   package Vec_Num is new Vec.Num;
   use Vec, Vec_Num;
   
   package TV is new Vectors(Scalar => Taylor, Zero_Scalar => ZeroFun(Radius'Last));
   
   subtype TayVector is TV.Vector;
   subtype TayMatrix is TV.Matrix;
   
   function DD(N: Integer) return Integer;
   procedure PrepNewt(N: in Integer; Ce: in Scalar; Eps,Rd,Rdz,Rdp1,Rp1,Rp11: in Radius;  S, Z: in Taylor; D: out Matrix; V: out Vector);
   --D is derivative of R(S)-S, and V is the vector S-R(S) in the symmetric basis.
   --Everything truncated to order N
   
end Doubling;
