with Reps;
use Reps;

package Intervals is

  pragma Elaborate_Body;

  type Interval is private;

--- basic
  function IsNumeric(S: Interval) return Boolean;
  function IsNumbers(S: Interval) return Boolean;
  function Err0(S: Interval) return Boolean;
  function Scal(K: Integer) return Interval;
  function Scal(R: Rep) return Interval;
  function Approx(S: Interval) return Rep;

  pragma Inline (IsNumeric,Err0,Scal,Approx);

--- sets
  procedure ModNumbers(S: in out Interval);
  function Ball(R: Rep) return Interval;
  function Ball(R,E: Rep) return Interval;
  function Ball(S: Interval) return Interval;
  function Disk(S: Interval) return Interval;
  function MaxAbs(S: Interval) return Radius;
  function Scal(R1,R2: Rep) return Interval;
  function Center(S: Interval) return Interval;
  function Contains(S: Interval; Z: Rep) return Boolean;
  procedure ResetCenter(S: in out Interval);
  function ResetCenter(S: Interval) return Interval;
  function ResetBall(S: Interval) return Interval;
  procedure Split(S: in Interval; S0,SE: out Interval);
  function Widen(S: Interval) return Interval;
  function Union(S1,S2: Interval) return Interval;

  pragma Inline (Ball,Center,ResetCenter,Split);

--- order
  function Inf(S: Interval) return Rep;
  function Sup(S: Interval) return Rep;
  function "<"(S1,S2: Interval) return Boolean;
  function "<="(S1,S2: Interval) return Boolean;
  function ">"(S1,S2: Interval) return Boolean;
  function ">="(S1,S2: Interval) return Boolean;
  function Min(S1,S2: Interval) return Interval;
  function Max(S1,S2: Interval) return Interval;

  pragma Inline (Inf,Sup);

--- ops
  function Norm(S: Interval) return Interval;
  function "abs"(S: Interval) return Interval renames Norm;
  function InfAbs(S: Interval) return Rep;
  function SupAbs(S: Interval) return Rep;
  function "-"(S: Interval) return Interval;
  function "+"(S,T: Interval) return Interval;
  function "-"(S,T: Interval) return Interval;
  function "*"(R: Rep; S: Interval) return Interval;
  procedure ErrMult(R: in Radius; S: in out Interval);
  function "/"(S: Interval; R: Rep) return Interval;
  function Sqr(S: Interval) return Interval;
  function "*"(S,T: Interval) return Interval;
  procedure MultAdd(S1,S2: in Interval; S3: in out Interval);
  function Inv(S: Interval) return Interval;
  function "/"(S,T: Interval) return Interval;
  function "**"(S: Interval; I: Integer) return Interval;

  pragma Inline (InfAbs,SupAbs,ErrMult,Norm,"abs","+","-","*",Sqr,Inv);

--- fun
  function Cosh(S: Interval) return Interval;
  function Sinh(S: Interval) return Interval;
  function Cos(S: Interval) return Interval;
  function Sin(S: Interval) return Interval;
  function Exp(S: Interval) return Interval;
  function Log(S: Interval) return Interval;
  function Sqrt(S: Interval) return Interval;
  function TruncExp(N: Natural; S: Interval) return Interval;
  function RemExp(N: Natural; S: Interval) return Interval;

private

   type Interval is  -- ball around constant
      record
         C: Rep;       -- center of ball
         R: Rep;       -- radius of ball
         E: Rep;       -- general radius, used for Taylors
      end record;

   procedure CoshSinh(S: in Interval; Sc,Ss: out Interval);
  function ExpUp(R: Rep) return Rep;
  function ExpDown(R: Rep) return Rep;

end Intervals;
