with Reps, ScalMisc, Vectors;
use Reps;

generic

  type Scalar is private;

  with function S_IsNumeric(S: Scalar) return Boolean is <>;
  with function S_IsNumbers(S: Scalar) return Boolean is <>;
  with function S_Err0(S: Scalar) return Boolean is <>;
  with function S_Scal(K: Integer) return Scalar is <>;
  with function S_Scal(R: Rep) return Scalar is <>;
  with function S_Approx(S: Scalar) return Rep is <>;

  with procedure S_ModNumbers(S: in out Scalar) is <>;
  with function S_Ball(R: Rep) return Scalar is <>;
  with function S_Ball(R,E: Rep) return Scalar is <>;
  with function S_Ball(S: Scalar) return Scalar is <>;
  with function S_Disk(S: Scalar) return Scalar is <>;
  with function S_MaxAbs(S: Scalar) return Radius is <>;
  with function S_Scal(R1,R2: Rep) return Scalar is <>;
  with function S_Center(S: Scalar) return Scalar is <>;
  with function S_Contains(S: Scalar; Z: Rep) return Boolean is <>;
  with procedure S_ResetCenter(S: in out Scalar) is <>;
  with function S_ResetCenter(S: Scalar) return Scalar is <>;
  with function S_ResetBall(S: Scalar) return Scalar is <>;
  with procedure S_Split(S: in Scalar; S0,SE: out Scalar) is <>;
  with function S_Widen(S: Scalar) return Scalar is <>;
  with function S_Union(S1,S2: Scalar) return Scalar is <>;

  with function S_Inf(S: Scalar) return Rep is <>;
  with function S_Sup(S: Scalar) return Rep is <>;
  with function S_LT(S1,S2: Scalar) return Boolean is <>;
  with function S_LE(S1,S2: Scalar) return Boolean is <>;
  with function S_GE(S1,S2: Scalar) return Boolean is <>;
  with function S_GT(S1,S2: Scalar) return Boolean is <>;
  with function S_Min(S1,S2: Scalar) return Scalar is <>;
  with function S_Max(S1,S2: Scalar) return Scalar is <>;

  with function S_Norm(S: Scalar) return Scalar is <>;
  with function S_InfAbs(S: Scalar) return Rep is <>;
  with function S_SupAbs(S: Scalar) return Rep is <>;
  with function S_Neg(S: Scalar) return Scalar is <>;
  with function S_Sum(S,T: Scalar) return Scalar is <>;
  with function S_Diff(S,T: Scalar) return Scalar is <>;
  with function S_Prod(R: Rep; S: Scalar) return Scalar is <>;
  with procedure S_ErrMult(R: in Rep; S: in out Scalar) is <>;
  with function S_Quot(S: Scalar; R: Rep) return Scalar is <>;
  with function S_Sqr(S: Scalar) return Scalar is <>;
  with function S_Prod(S,T: Scalar) return Scalar is <>;
--  with procedure S_MultAdd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function S_Inv(S: Scalar) return Scalar is <>;
  with function S_Quot(S,T: Scalar) return Scalar is <>;
  with function S_Pow(S: Scalar; I: Integer) return Scalar is <>;

  with function S_Cosh(S: Scalar) return Scalar is <>;
  with function S_Sinh(S: Scalar) return Scalar is <>;
  with function S_Cos(S: Scalar) return Scalar is <>;
  with function S_Sin(S: Scalar) return Scalar is <>;
  with function S_Exp(S: Scalar) return Scalar is <>;
  with function S_Log(S: Scalar) return Scalar is <>;
  with function S_Sqrt(S: Scalar) return Scalar is <>;
  with function S_TruncExp(N: Natural; S: Scalar) return Scalar is <>;
  with function S_RemExp(N: Natural; S: Scalar) return Scalar is <>;

package Scalars is

 --- basic
  function IsNumeric(S: Scalar) return Boolean renames S_IsNumeric;
  function IsNumbers(S: Scalar) return Boolean renames S_IsNumbers;
  function Err0(S: Scalar) return Boolean renames S_Err0;
  function Scal(K: Integer) return Scalar renames S_Scal;
  function Scal(R: Rep) return Scalar renames S_Scal;
  function Approx(S: Scalar) return Rep renames S_Approx;

  pragma Inline (IsNumeric,Err0,Scal,Approx);

  --- sets
  procedure  ModNumbers(S: in out Scalar) renames S_ModNumbers;
  function Ball(R: Rep) return Scalar renames S_Ball;
  function Ball(R,E: Rep) return Scalar renames S_Ball;
  function Ball(S: Scalar) return Scalar renames S_Ball;
  function Disk(S: Scalar) return Scalar renames  S_Disk;
  function MaxAbs(S: Scalar) return Radius renames S_MaxAbs;
  function Scal(R1,R2: Rep) return Scalar renames S_Scal;
  function Center(S: Scalar) return Scalar renames S_Center;
  function Contains(S: Scalar; Z: Rep) return Boolean renames S_Contains;
  procedure ResetCenter(S: in out Scalar) renames S_ResetCenter;
  function ResetCenter(S: Scalar) return Scalar renames S_ResetCenter;
  function ResetBall(S: Scalar) return Scalar renames S_ResetBall;
  procedure Split(S: in Scalar; S0,SE: out Scalar) renames S_Split;
  function Widen(S: Scalar) return Scalar renames S_Widen;
  function Union(S1,S2: Scalar) return Scalar renames S_Union;

  pragma Inline (Ball,Scal,Center,ResetCenter,ResetBall,Split);

--- order
  function Inf(S: Scalar) return Rep renames S_Inf;
  function Sup(S: Scalar) return Rep renames S_Sup;
  function "<"(S1,S2: Scalar) return Boolean renames S_LT;
  function "<="(S1,S2: Scalar) return Boolean renames S_LE;
  function ">"(S1,S2: Scalar) return Boolean renames S_GT;
  function ">="(S1,S2: Scalar) return Boolean renames S_GE;
  function Min(S1,S2: Scalar) return Scalar renames S_Min;
  function Max(S1,S2: Scalar) return Scalar renames S_Max;

  pragma Inline (Inf,Sup);

--- ops
  function Norm(S: Scalar) return Scalar renames S_Norm;
  function "abs"(S: Scalar) return Scalar renames S_Norm;
  function InfAbs(S: Scalar) return Rep renames S_InfAbs;
  function SupAbs(S: Scalar) return Rep renames S_SupAbs;
  function "-"(S: Scalar) return Scalar renames S_Neg;
  function Sum(S,T: Scalar) return Scalar renames S_Sum;
  function "+"(S,T: Scalar) return Scalar renames S_Sum;
  function Diff(S,T: Scalar) return Scalar renames S_Diff;
  function "-"(S,T: Scalar) return Scalar renames S_Diff;
  function "*"(R: Rep; S: Scalar) return Scalar renames S_Prod;
  procedure ErrMult(R: in Rep; S: in out Scalar) renames S_ErrMult;
  function "/"(S: Scalar; R: Rep) return Scalar renames S_Quot;
  function Sqr(S: Scalar) return Scalar renames S_Sqr;
  function Prod(S,T: Scalar) return Scalar renames S_Prod;
  function "*"(S,T: Scalar) return Scalar renames S_Prod;
--  procedure MultAdd(S1,S2: in Scalar; S3: in out Scalar) renames S_MultAdd;
  function Inv(S: Scalar) return Scalar renames S_Inv;
  function Quot(S,T: Scalar) return Scalar renames S_Quot;
  function "/"(S,T: Scalar) return Scalar renames S_Quot;
  function "**"(S: Scalar; I: Integer) return Scalar renames S_Pow;
--  function Pow(S: Scalar; I: Integer) return Scalar renames S_Pow;


  pragma Inline (InfAbs,SupAbs,ErrMult,Norm,"abs",Sum,"+",Diff,"-",Prod,"*",Sqr,Quot,"/",Inv);

--- fun
  function Cosh(S: Scalar) return Scalar renames S_Cosh;
  function Sinh(S: Scalar) return Scalar renames S_Sinh;
  function Cos(S: Scalar) return Scalar renames S_Cos;
  function Sin(S: Scalar) return Scalar renames S_Sin;
  function Exp(S: Scalar) return Scalar renames S_Exp;
  function Log(S: Scalar) return Scalar renames S_Log;
  function Sqrt(S: Scalar) return Scalar renames S_Sqrt;
  function TruncExp(N: Natural; S: Scalar) return Scalar renames S_TruncExp;
  function RemExp(N: Natural; S: Scalar) return Scalar renames S_RemExp;

--- misc
  Zero_Scalar: constant Scalar := Scal(Zero);

  package SM is new ScalMisc (Scalar => Scalar, Zero_Scalar => Zero_Scalar);

  function IsScalar(S: Scalar) return Boolean renames SM.IsScalar;
  function IsZero(S: Scalar) return Boolean renames SM.IsZero;
  procedure Neg(S: in out Scalar) renames SM.Neg;
  procedure Sum(P,Q: in Scalar; F: out Scalar) renames SM.Sum;
  procedure Add(S: in Scalar; T: in out Scalar) renames SM.Add;
  procedure Diff(P,Q: in Scalar; F: out Scalar) renames SM.Diff;
  procedure Sub(S: in Scalar; T: in out Scalar) renames SM.Sub;
  procedure Prod(S: in Scalar; P: in Scalar; Q: out Scalar) renames SM.Prod;
  procedure Mult(S: in Scalar; T: in out Scalar) renames SM.Mult;
  procedure MultAdd(R,S: in Scalar; T: in out Scalar) renames SM.MultAdd;
  procedure Div(S: in Scalar; T: in out Scalar) renames SM.Div;

  pragma Inline (Neg,Add,Sub,Mult,MultAdd,Div);

  package SV is new Vectors(Scalar => Scalar, Zero_Scalar => Zero_Scalar);

  subtype ScalVec is SV.Vector;
  subtype ScalMat is SV.Matrix;
  subtype ScalPoly1 is SV.Polynom1;
  subtype ScalPoly2 is SV.Polynom2;

end Scalars;
