with Messages;
use Messages;

package body Taylors is

   Zero_Taylor: Taylor;

   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;

   function IsNumbers(F: Taylor) return Boolean is
   begin
      if not Trunc then
	 for I in Power loop
	    for J in 0 .. PDeg-I loop
	       if not IsNumbers(F.C(I,J)) then return False; end if;
	    end loop;
	 end loop;
      end if;
      return True;
   end IsNumbers;


   function InDomain(R: Radius; F: Taylor) return Boolean is
   begin
      return (R<= F.R ) or else IsNumbers(F);
   end InDomain;


   function Rho(F: Taylor) return Radius is
   begin
      return F.R;
   end Rho;
   
   
   procedure SetRho(R: in Radius; F: out Taylor) is
   begin
      F.R := R;
   end SetRho;


   function IsZero(F: Taylor) return Boolean is
   begin
      return F.C(1,1)=ScalZero and then (F.C=Zero_Taylor.C);
   end IsZero;


   procedure SetZero(R: in Radius; F: out Taylor) is
   begin
      F := Zero_Taylor;
      F.R := R;
   end SetZero;
   
   
   procedure ResetErrComp(F: in out Taylor) is
   begin
      for I in Power loop
	 for J in Power loop
	    F.C(I,J):=Center(F.C(I,J));
	 end loop;
      end loop;
   end ResetErrComp;


   function EffDegg(D: Integer; P: Taylor) return Integer is --fine
   begin
      for I in reverse 0..D loop
	 for J in reverse 0 .. D-I loop
	    if P.C(I,J) /= ScalZero then return I;
	    end if;
	 end loop;
      end loop;
      --  if Trunc or else P.E=Zero then return -1; end if;
      return -1;
   end EffDegg;


   function EffDeg(P: Taylor) return Integer is --fine
   begin
      for J in reverse Power loop
	 for I in reverse 0 .. J loop
	    if P.C(I,J-I) /= ScalZero then return J;
	    end if;
	 end loop;
      end loop;
      --  if Trunc or else P.E=Zero then return -1; end if;
      return -1;
   end EffDeg;
   
   
   function ZeroFun(R: Radius) return Taylor is
      F: Taylor := Zero_Taylor;
   begin
      F.R := R;
      return F;
   end ZeroFun;
   
   
   function Monom(R: Radius; I,J: Natural; S: Scalar) return Taylor is
      F: Taylor := Zero_Taylor;
   begin
      F.R := R;
      F.C(I,J) := S;
      return F;
   end Monom;
   
   
   function Component(I,J: Natural; F: Taylor) return Scalar is
   begin
      return F.C(I,J);
   end Component;
   
   
   procedure ZeroComponent(I,J: in  Natural; F: in out Taylor) is
   begin
      F.C(I,J):=Scal(Zero);
   end ZeroComponent;
   
   
   function Component(F,H: Taylor) return Scalar is
      S: Scalar:=Scal(Zero);
      C: Scalar:=Scal(Zero);
   begin
      for I in Power loop
	 for J in Power loop
	    S:=S+F.C(I,J)*H.C(I,J);
	    C:=C+Sqr(H.C(I,J));
	 end loop;
      end loop;
      return S/Sqrt(C);
   end Component;
   
   
   procedure Neg(F: in out Taylor) is
   begin
      Neg(F.C);
   end Neg;
   
   
   function "-"(F: Taylor) return Taylor is
   begin
      return (F.R,-F.C);
   end "-";
   
   
   procedure Mult(S: in Scalar; F: in out Taylor) is
   begin
      Mult(S,F.C);
   end Mult;
   
   
   function "*"(S: Scalar; F: Taylor) return Taylor is
   begin
      return (F.R,S*F.C);
   end "*";
   
   
   procedure Add(F1: in Taylor; F2: in out Taylor) is
   begin
      F2.R := Radius'Min(F1.R,F2.R);
      Add(F1.C,F2.C);
   end Add;
   
   
   function "+"(F1,F2: Taylor) return Taylor is
      F3: Taylor;
   begin
      F3.R := Radius'Min(F1.R,F2.R);
      F3.C := F1.C+F2.C;
      return F3;
   end "+";
   
   
   procedure Sub(F1: in Taylor; F2: in out Taylor) is
   begin
      F2.R := Radius'Min(F1.R,F2.R);
      Sub(F1.C,F2.C);
   end Sub;
   
   
   function "-"(F1,F2: Taylor) return Taylor is
      F3: Taylor;
   begin
      F3.R := Radius'Min(F1.R,F2.R);
      F3.C := F1.C-F2.C;
      return F3;
   end "-";


   procedure Combine(R: in Radius; DF: in Natural; F: in out  Matrix;  G: in out Poly2) is
      --- include higher order terms
   begin
      if DF <= PDeg then return; end if;
      for D in reverse PDeg .. DF-1 loop
	 F(0,D):=R*F(0,D+1)+F(0,D);
	 for I in 0 .. D loop
	    F(I,D-I):=R*F(I+1,D-I)+F(I,D-I);
	 end loop;
      end loop;
      for I in Power loop
	 G(I,PDeg-I):=Ball(Zero,MaxAbs(F(I,PDeg-I)))+G(I,PDeg-I);
      end loop;
   end Combine;


   procedure MultAdd(DF,DG: in Natural; F,G: in Taylor; H: in out Taylor) is
      --- only adds terms of order PDeg or less
      DF1:  constant Integer := EffDegg(DF,F);
      DG1:  constant Integer := EffDegg(DG,G);
      DFG:  constant Integer := IMin(DF+DG,PDeg);
      DFG1: constant Integer := IMin(DF1+DG1,PDeg);
      S: Scalar;
   begin
      for IH in reverse 0 .. DFG1 loop
	 for JH in 0 .. DFG-IH loop
	    S := ScalZero;
	    for I in Imax(IH-DG1,0) .. IMin(IH,DF1) loop
	       for J in Imax(JH-DG,0) .. IMin(JH,DF) loop
		  S:=F.C(I,J)*G.C(IH-I,JH-J)+S;
	       end loop;
	    end loop;
	    H.C(IH,JH):=S+H.C(IH,JH);
	 end loop;
      end loop;
   end MultAdd;


   procedure MultAdd(F,G: in Taylor; H: in out Taylor) is
      DF:   constant Integer := EffDeg(F);
      DG:   constant Integer := EffDeg(G);
      DFG:  constant Integer := DF+DG;
   begin
      if DF<0 or else DG<0 then return; end if;
      H.R := Rep'Min(Rep'Min(F.R,G.R),H.R);
      if Trunc or else DFG <= PDeg then          -- no higher order terms
	 MultAdd(DF,DG,F,G,H);
	 return;
      end if;
      declare
	 DF1:  constant Integer := EffDegg(DF,F);
	 DG1:  constant Integer := EffDegg(DG,G);
	 DFG1: constant Integer := DF1+DG1;
	 S: Scalar;
	 Higher: Matrix(0..2*PDeg,0..2*PDeg);
      begin
	 SetZero(Higher);
	 for IH in reverse 0 .. DFG1 loop
	    for JH in 0 .. DFG-IH loop
	       S := ScalZero;
	       for I in Imax(IH-DG1,0) .. IMin(IH,DF1) loop
		  for J in Imax(JH-DG,0) .. IMin(JH,DF) loop
		     S:=F.C(I,J)*G.C(IH-I,JH-J)+S;
		  end loop;
	       end loop;
	       if IH+JH <= PDeg then
		  H.C(IH,JH):=S+H.C(IH,JH);
	       else
		  Higher(IH,JH) := S;
	       end if;
	    end loop;
	 end loop;
	 Combine(H.R,DFG,Higher,H.C);
      end;
   end MultAdd;


   procedure MultAdd(S: in Scalar; F: in Taylor; G: in out Taylor) is
   begin
      if S /= ScalZero then
	 G.R := Rep'Min(F.R,G.R);
	 for I in Power loop
	    for J in 0 .. PDeg-I loop
	       G.C(I,J):=S*F.C(I,J)+G.C(I,J);
	    end loop;
	 end loop;
      end if;
   end MultAdd;


   procedure Mult(F: in Taylor; G: in out Taylor) is
      H: constant Taylor := G;
   begin
      SetZero(Rho(H),G);
      MultAdd(F,H,G);
   end Mult;


   function Val(S1,S2: Scalar; F: Taylor) return Scalar is
      R: constant Radius :=  Rep'Max(MaxAbs(S1),MaxAbs(S2));
      S3,S4: Scalar;
   begin
      if not InDomain(R,F) then
	 Message("Taylors.Val",Domain_Error);
      end if;
      S3 := ScalZero;
      for I in reverse Power loop
	 S4 := ScalZero;
	 for J in reverse 0 .. PDeg-I loop
	    S4:=S2*S4;
	    S4:=F.C(I,J)+S4;
	 end loop;
	 S3:=S1*S3;
	 S3:=S4+S3;
      end loop;
      return Disk(S3);
   end Val;


   function Val(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar is
      T: Scalar := Val(S1,S2,F.C);
   begin
      T:=(C+E)*(S1-Scal(Half))+E*(S2+Scal(Half))+Half*B*Sqr(S2+Scal(Half))+T;
      return T;
   end Val;


   procedure PolyDer1(F: in Taylor; G: out Taylor) is
   begin
      SetZero(F.R,G);
      for I in 1 .. PDeg loop
	 for J in 0 .. PDeg-I loop
	    G.C(I-1,J):=Rep(I)*F.C(I,J);
	 end loop;
      end loop;
   end PolyDer1;
   
   
   procedure PolyDer2(F: in Taylor; G: out Taylor) is
   begin
      SetZero(F.R,G);
      for J in 1 .. PDeg loop
	 for I in 0 .. PDeg-J loop
	    G.C(I,J-1):=Rep(J)*F.C(I,J);
	 end loop;
      end loop;
   end PolyDer2;
   
   
   function DerFac1(R1,R2: Radius) return Scalar is
      --- norm factor for 1st derivative
      --- R1: old domain, R2: new domain
      M: constant Scalar := Inv(Log(Scal(R1)/R2));
   begin
      if R1 <= R2 then
	 Message("Taylors.DerFac1",Domain_Error);
      end if;
      if M<ScalOne then
	 return ScalOne/R1;
      else
	 return (M/R2)*Exp(-ScalOne);
      end if;
   end DerFac1;
   
   
   function DerFac2(R1,R2: Radius) return Scalar is
      --- norm factor for 2nd derivative
      --- R1: old domain, R2: new domain
      L: constant Scalar := Log(Scal(R1)/R2);
      M: constant Scalar := (L+ScalTwo+Sqrt(L*L+Scal(Rep(4.0))))/(Two*L);
   begin
      if R1 <= R2 then
	 Message("Taylors.DerFac2",Domain_Error);
      end if;
      if M<ScalTwo then
	 return Two*Sqr(ScalOne/R1);
      else
	 return (M/R2)*((M-ScalOne)/R2)*Exp(-(Two*M-ScalOne)/(M-ScalOne));
      end if;
   end DerFac2;
   
   
   procedure Ders(R: in Radius; F0: in Taylor; F: out Taylor) is
   begin
      F := F0;
      SetRho(R,F);
   end Ders;


   procedure ModNumbers(F: in out Taylor) is
   begin
      if Trunc then
	 F := Zero_Taylor;
      else
	 for I in Power loop
	    for J in 0 .. PDeg-I loop
	       ModNumbers(F.C(I,J));
	    end loop;
	 end loop;
      end if;
   end ModNumbers;


   procedure ModNumbers(F: in Taylor; G: out Taylor) is
   begin
      if Trunc then
	 SetZero(F.R,G);
      else
	 G := F;
	 for I in Power loop
	    for J in 0 .. PDeg-I loop
	       ModNumbers(G.C(I,J));
	    end loop;
	 end loop;
      end if;
   end ModNumbers;


   function ModNumbers(F: Taylor) return Taylor is
      G: Taylor;
   begin
      ModNumbers(F,G);
      return G;
   end ModNumbers;


   procedure Der1(R: in Radius; F0: in Taylor; Fu: out Taylor) is
      F: Taylor;
   begin
      Ders(R,F0,F);
      PolyDer1(F,Fu);
      if IsNumbers(F) then return; end if;
      declare
	 U: constant Scalar := DerFac1(F0.R,R);
	 G: Taylor;
      begin
	 ModNumbers(F,G);
	 MultAdd(U,G,Fu);
      end;
   end Der1;


   procedure Der2(R: in Radius; F0: in Taylor; Fv: out Taylor) is
      F: Taylor;
   begin
      Ders(R,F0,F);
      PolyDer2(F,Fv);
      if IsNumbers(F) then return; end if;
      declare
	 U: constant Scalar := DerFac1(F0.R,R);
	 G: Taylor;
      begin
	 ModNumbers(F,G);
	 MultAdd(U,G,Fv);
      end;
   end Der2;


   procedure Ders(R: in Radius; F0: in Taylor; Fu,Fv: out Taylor) is
      F: Taylor;
   begin
      Ders(R,F0,F);
      PolyDer1(F,Fu);
      PolyDer2(F,Fv);
      if IsNumbers(F) then return; end if;
      declare
	 U: constant Scalar := DerFac1(F0.R,R);
	 G: Taylor;
      begin
	 ModNumbers(F,G);
	 MultAdd(U,G,Fu);
	 MultAdd(U,G,Fv);
      end;
   end Ders;


   function Der1(R: Radius; F: Taylor) return Taylor is
      T: Taylor;
   begin
      Der1(R,F,T);
      return T;
   end Der1;
   
   
   function Der2(R: Radius; F: Taylor) return Taylor is
      T: Taylor;
   begin
      Der2(R,F,T);
      return T;
   end Der2;


   procedure DerDers(R: in Radius; F0: in Taylor; F,Fu,Fv,Fuu,Fuv,Fvv: out Taylor) is
   begin
      Ders(R,F0,F);
      PolyDer1(F,Fu);
      PolyDer2(F,Fv);
      PolyDer1(Fu,Fuu);
      PolyDer2(Fu,Fuv);
      PolyDer2(Fv,Fvv);
      if IsNumbers(F) then return; end if;
      declare
	 U1: constant Scalar := DerFac1(F0.R,R);
	 U2: constant Scalar := DerFac2(F0.R,R);
	 G: Taylor;
      begin
	 ModNumbers(Fu,G);
	 MultAdd(U1,G,Fuv);
	 MultAdd(Two*U1,G,Fuu);
	 ModNumbers(Fv,G);
	 MultAdd(U1,G,Fuv);
	 MultAdd(Two*U1,G,Fvv);
	 ModNumbers(F,G);
	 MultAdd(U1,G,Fu);
	 MultAdd(U1,G,Fv);
	 MultAdd(U2,G,Fuu);
	 MultAdd(U2,G,Fvv);
	 MultAdd(U1*U1,G,Fuv);
      end;
   end DerDers;


   procedure DerDer22(R: in Radius; F0: in Taylor; Fvv: out Taylor) is
      F,Fv: Taylor;
   begin
      Ders(R,F0,F);
      PolyDer2(F,Fv);
      PolyDer2(Fv,Fvv);
      if IsNumbers(F) then return; end if;
      declare
	 U1: constant Scalar := DerFac1(F0.R,R);
	 U2: constant Scalar := DerFac2(F0.R,R);
	 G: Taylor;
      begin
	 ModNumbers(Fv,G);
	 MultAdd(Two*U1,G,Fvv);
	 ModNumbers(F,G);
	 MultAdd(U2,G,Fvv);
      end;
   end DerDer22;


   procedure DerDer11(R: in Radius; F0: in Taylor; Fuu: out Taylor) is
      F, Fu: Taylor;
   begin
      Ders(R,F0,F);
      PolyDer1(F,Fu);
      PolyDer1(Fu,Fuu);
      if IsNumbers(F) then return; end if;
      declare
	 U1: constant Scalar := DerFac1(F0.R,R);
	 U2: constant Scalar := DerFac2(F0.R,R);
	 G: Taylor;
      begin
	 ModNumbers(Fu,G);
	 MultAdd(Two*U1,G,Fuu);
	 ModNumbers(F,G);
	 MultAdd(U2,G,Fuu);
      end;
   end DerDer11;


   procedure DerDer12(R: in Radius; F0: in Taylor; Fuv: out Taylor) is
      F,Fu,Fv: Taylor;
   begin
      Ders(R,F0,F);
      PolyDer1(F,Fu);
      PolyDer2(F,Fv);
      PolyDer2(Fu,Fuv);
      if IsNumbers(F) then return; end if;
      declare
	 U1: constant Scalar := DerFac1(F0.R,R);
	 U2: constant Scalar := DerFac2(F0.R,R);
	 G: Taylor;
      begin
	 ModNumbers(Fu,G);
	 MultAdd(U1,G,Fuv);

	 ModNumbers(Fv,G);
	 MultAdd(U1,G,Fuv);

	 ModNumbers(F,G);
	 MultAdd(U1*U1,G,Fuv);
      end;
   end DerDer12;


   function DerDer11(R1: Radius; F: Taylor) return Taylor is
      T: Taylor;
   begin
      DerDer11(R1,F,T);
      return T;
   end DerDer11;


   function DerDer22(R1: Radius; F: Taylor) return Taylor is
      T: Taylor;
   begin
      DerDer22(R1,F,T);
      return T;
   end DerDer22;


   function DerDer12(R1: Radius; F: Taylor) return Taylor is
      T: Taylor;
   begin
      DerDer12(R1,F,T);
      return T;
   end DerDer12;


   function DerVal1(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar is
      R1: Rep:=Rep(1.000001)*Rep'Max(MaxAbs(S1),MaxAbs(S2));
      F1: Taylor:=Der1(R1,F);
      T: Scalar :=Val(S1,S2,F1);
   begin
      return (C+E)+T;
   end DerVal1;
   
   
   function DerVal2(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar is
      R1: Rep:=Rep(1.000001)*Rep'Max(MaxAbs(S1),MaxAbs(S2));
      F1: Taylor:=Der2(R1,F);
      T: Scalar :=Val(S1,S2,F1);
   begin
      return E+B*(S2+Scal(Half))+T;
   end DerVal2;

   --------------------------------------------------

   function "*"(F1,F2: Taylor) return Taylor is
      F3: Taylor := F2;
   begin
      Mult(F1,F3);
      return F3;
   end "*";


   function Inv(F: Taylor) return Taylor is
      C: constant Scalar := Inv(F.C(0,0));
      Ft: Taylor := (-C)*F;
      G,G1: Taylor;
      R: Radius;
      Eps: Rep;
      Der,B: Scalar;
   begin
      R:=Rho(F);
      SetZero(R,G1);
      SetZero(R,G);

      Ft.C(0,0) := ScalZero;
      G.C(0,0) := ScalOne;
      for I in reverse 0 .. 100 loop
	 G1:=G;
	 Mult(Ft,G);
	 G.C(0,0) := G.C(0,0)+ScalOne;
	 Eps:=MaxAbs(Norm(R,G-G1));
	 if Eps<Rep(1.0E-17) then exit; end if;
      end loop;

      if Trunc then return C*G; end if;

      Der:=Norm(Rho(Ft),Ft);
      B:=Ball(Zero,MaxAbs(Eps*Inv(Scal(One)-Der)));
      G:=G+Monom(Rho(G),0,0,B);
      return C*G;
   end Inv;


   function Sqrt(H: Taylor) return Taylor is
      R: Radius;
      S: Rep;
      T,TT,HH: Taylor;
      C,RR,Nr,Nh,C0: Scalar;
      NF,TNF,K: Integer;
      N: Integer:=30;
   begin
      HH:=H;
      C0:=Scal(One);
      C0:=Component(0,0,HH);
      ZeroComponent(0,0,HH);
      C0:=(Scal(One)+C0);
      HH:=Inv(C0)*HH;
      C0:=Sqrt(C0);


      R:=Rho(HH);
      Nh:=Norm(R,HH);
      if MaxAbs(Nh) >=One then
	 Message("Taylors.Sqrt",Singular);
      end if;

      SetZero(R,T);
      T:=T+Monom(R,0,0,Scal(One));

      S:=-One;
      NF:=1;
      TNF:=1;
      C:=Scal(One);
      SetZero(R,TT);
      TT:=TT+Monom(R,0,0,Scal(One));
      for I in 1..N loop
	 K:=2*I;
	 TNF:=TNF*(K-1)*K;
	 NF:=NF*I;
	 TT:=TT*HH;
	 C:=C*Scal(4.0);
	 T:=T+(S*(Scal(TNF)/((Scal(One)-Scal(K))*Sqr(Scal(NF))*C)))*TT;
	 S:=-S;
      end loop;
      if Trunc then return C0*T; end if;

      K:=2*(N+1);
      TNF:=TNF*(K-1)*K;
      NF:=NF*(N+1);
      TT:=TT*HH;
      C:=C*Scal(4.0);

      Nr:=Nh**(N+1);
      RR:=(Scal(R))**(N+1);
      --T:=T+Monom(R,PDeg,PDeg,Ball(Zero,MaxAbs(((Scal(TNF)/((Scal(One)-Scal(K))*Sqr(Scal(NF))*C)))*Nr/(RR*(Scal(One)-Nh)))));
      T:=T+Monom(R,N+1,0,Ball(Zero,MaxAbs(((Scal(TNF)/((Scal(One)-Scal(K))*Sqr(Scal(NF))*C)))*Nr/(RR*(Scal(One)-Nh)))));
      T:=C0*T;
      return T;
   end Sqrt;


   function "/"(F1,F2: Taylor) return Taylor is
   begin
      return F1*Inv(F2);
   end "/";


   procedure Truncate(N1,N2: in Integer; F: in out  Taylor) is
   begin
      for I in Power loop
	 for J in Power loop
	    if I>N1 or J>N2 then
	       F.C(I,J):=Scal(Zero);
	    end if;
	 end loop;
      end loop;
   end Truncate;


   function Norm(R: Radius; F: Taylor) return Scalar is
      S1,S2: Scalar;
   begin
      S1 := ScalZero;
      if ( (not Trunc) and (not InDomain(R,F)) ) then
	 Message("Taylors.Norm",Domain_Error);
      end if;
      for I in reverse Power loop
	 S2 := ScalZero;
	 for J in reverse 0 .. PDeg-I loop
	    S2 := R*S2+Norm(F.C(I,J));
	 end loop;
	 S1 := R*S1+S2;
      end loop;
      return Disk(S1);
   end Norm;


   function Norm(F: Taylor) return Scalar is
   begin
      return Norm(F.R,F);
   end Norm;

   function Invert(R: Rep; F,InvF0: Taylor) return Taylor is
      InvF: Taylor:=InvF0;
   begin
      for I in 1..200 loop
	 InvF:=InvF+(Monom(R,0,1,Scal(One))-Comp2(InvF,F) )/Comp2(InvF,Der2(R,F));
      end loop;
      return InvF;
   end Invert;


   function Comp1(F,H: Taylor) return Taylor is
      R: constant Radius :=F.R;
      G1,G2: Taylor;
   begin
      if ( (not Trunc) and (MaxAbs(Norm(F.R,F))>H.R) )  then
	 Message("Taylors.Comp1",Domain_Error);
      end if;

      SetZero(R,G1);
      for I in reverse Power loop
	 SetZero(R,G2);
	 for J in reverse Power loop
	    G2.C(0,J) := H.C(I,J);
	 end loop;
	 --MultAdd(F,G1,G2);
	 Mult(F,G1);
	 Add(G2,G1);
      end loop;

      return G1;
   end Comp1;
   
   
   function Comp2(F,H: Taylor) return Taylor is
      R: constant Radius := F.R;
      G1,G2: Taylor;
   begin
      if ( (not Trunc) and (MaxAbs(Norm(F.R,F))>H.R) ) then
	 Message("Taylors.Comp2",Domain_Error);
      end if;

      SetZero(R,G1);
      for J in reverse Power loop
	 SetZero(R,G2);
	 for I in reverse Power loop
	    G2.C(I,0) := H.C(I,J);
	 end loop;

	 Mult(F,G1);
	 Add(G2,G1);
	 --MultAdd(F,G1,G2);
      end loop;
      return G1;
   end Comp2;
   
   
   function Comp(F1,F2,H: Taylor) return Taylor is
      R: constant Radius := Rep'Min(F1.R,F2.R);
      G1,G2: Taylor;
   begin

      if not Trunc then
	 if (MaxAbs(Norm(F1.R,F1))>H.R) or else  (MaxAbs(Norm(F2.R,F2))>H.R) then
	    Message("Taylors.Norm",Domain_Error);
	 end if;
      end if;
      SetZero(R,G1);
      for I in reverse Power loop
	 SetZero(R,G2);
	 for J in reverse Power loop
	    Mult(F2,G2);
	    G2.C(0,0) := G2.C(0,0)+H.C(I,J);
	 end loop;
	 Mult(F1,G1);
	 Add(G2,G1);
      end loop;

      return G1;
   end Comp;


   function SupRad(C: Scalar; Tay: Taylor) return Radius is
      R: Rep;
   begin
      R:=SupAbs(C);
      return Rep(1.000001)*R;
   end SupRad;


begin

   Zero_Taylor.R := Radius'Last;
   SetZero(Zero_Taylor.C);

end Taylors;
