with Reps, Scalars, Vectors;
use Reps;

generic

   PDeg: in Natural;
   CutOff: in Natural;
   with package Numbers is new Scalars (<>);

package Taylors is

  pragma Elaborate_Body;
  use Numbers;

  type Taylor is private;

  function InDomain(R: Radius; F: Taylor) return Boolean;
  function Rho(F: Taylor) return Radius;
  procedure SetRho(R: in Radius; F: out Taylor);
  function IsZero(F: Taylor) return Boolean;
  procedure SetZero(R: in Radius; F: out Taylor);
  procedure ResetErrComp(F: in out Taylor);
  function ZeroFun(R: Radius) return Taylor;
  function Monom(R: Radius; I,J: Natural; S: Scalar) return Taylor;
  function Component(I,J: Natural; F: Taylor) return Scalar;
  function Component(F,H: Taylor) return Scalar;
  procedure ZeroComponent(I,J: in  Natural; F: in out Taylor);
  procedure Neg(F: in out Taylor);
  function "-"(F: Taylor) return Taylor;
  procedure Mult(S: in Scalar; F: in out Taylor);
  function "*"(S: Scalar; F: Taylor) return Taylor;
  procedure Add(F1: in Taylor; F2: in out Taylor);
  function "+"(F1,F2: Taylor) return Taylor;
  procedure Sub(F1: in Taylor; F2: in out Taylor);
  function "-"(F1,F2: Taylor) return Taylor;
  procedure Mult(F: in Taylor; G: in out Taylor);
  function "*"(F1,F2: Taylor) return Taylor;
  procedure MultAdd(F,G: in Taylor; H: in out Taylor);
  function Inv(F: Taylor) return Taylor;
  function Sqrt(H: Taylor) return Taylor;
  function "/"(F1,F2: Taylor) return Taylor;

  function SupRad(C: Scalar; Tay: Taylor) return Radius;

  function Val(S1,S2: Scalar; F: Taylor) return Scalar;
  function Val(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar;
  function DerVal1(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar;
  function DerVal2(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar;
  procedure Truncate(N1,N2: in Integer; F: in out  Taylor);
  function Norm(R: Radius; F: Taylor) return Scalar;
--  function Norm(Rx,Ry: Radius; F: Taylor) return Scalar;
  function Norm(F: Taylor) return Scalar;

  procedure Der1(R: in Radius; F0: in Taylor; Fu: out Taylor);
  procedure Der2(R: in Radius; F0: in Taylor; Fv: out Taylor);
  procedure Ders(R: in Radius; F0: in Taylor; Fu,Fv: out Taylor);

  function  Der1(R: Radius; F: Taylor) return Taylor;
  function  Der2(R: Radius; F: Taylor) return Taylor;

  procedure DerDer11(R: in Radius; F0: in Taylor; Fuu: out Taylor);
  function DerDer11(R1: Radius; F: Taylor) return Taylor;
  procedure DerDer22(R: in Radius; F0: in Taylor; Fvv: out Taylor);
  function DerDer22(R1: Radius; F: Taylor) return Taylor;
  procedure DerDer12(R: in Radius; F0: in Taylor; Fuv: out Taylor);
  function DerDer12(R1: Radius; F: Taylor) return Taylor;

  function Comp1(F,H: Taylor) return Taylor;                        -- H(F,.)
  function Comp2(F,H: Taylor) return Taylor;                        -- H(.,F)
  function Comp(F1,F2,H: Taylor) return Taylor;                     -- H(F1,F2)

--  procedure Quot(P1,P2: in Taylor; P3: out Taylor);
--  function Quot(P1,P2: Taylor; D: Degree := Size) return Taylor;
--  function "/"(P1,P2: Taylor) return Taylor;



  --ROUTINES USED TO SOLVE THE PHI-EQUATION
  function Invert(R: Rep;F,InvF0: Taylor) return Taylor;

private

  ScalZero:  constant Scalar := Scal(Zero);
  ScalOne:  constant Scalar := Scal(One);
  ScalTwo:  constant Scalar := Scal(Two);
  ScalHalf:  constant Scalar := Scal(Half);
  Trunc: constant Boolean := IsNumeric(ScalZero);

  subtype Power is Integer range 0 .. PDeg;

  package SV is new Vectors(Scalar => Scalar, Zero_Scalar => ScalZero);
  use SV;

  subtype Poly2 is Polynom2(Power,Power);

  type Taylor is
    record
      R: Radius;      --- domain
      C: Poly2;       --- coefficients
    end record;

end Taylors;
