with Messages,Ada.Numerics.Generic_Elementary_Functions,Vectors.IO,Ada.Text_IO;
use Ada.Text_IO;
use Messages;

package body Vectors.Num is

   Zero_Scalar: constant Scalar := Scal(0);
   One_Scalar: constant Scalar := Scal(1);
   Two_Scalar: constant Scalar := Scal(2);

   function Norm2(B: Vector) return Scalar is
      R: Scalar;
   begin
      R := Zero_Scalar;
      for I in B'Range loop
         R := R+B(I)*B(I);
      end loop;
      return Sqrt(abs(R));
   end Norm2;
   
   
   procedure Decomp(A: in out Matrix; Cond: out Scalar; Pvt: out Pivot; Work: out Vector) is
      --- translated from numerical recipes
      N0,N,M: Integer;
      T,Anorm,Ynorm,Znorm: Scalar;
   begin
      SetZero(Work);
      N0 := Work'First;
      N := Work'Last;
      Pvt(N) := N0;
      if N = N0 then
         Cond := One_Scalar;
         if A(N0,N0) = Zero_Scalar then
            Message("Vectors.Decomp",Singular);
         end if;
	 return;
      end if;
      Anorm := Zero_Scalar;
      for J in N0 .. N loop
         T := Zero_Scalar;
         for I in N0 .. N loop
            T := T+abs(A(I,J));
         end loop;
         if T > Anorm then
            Anorm := T;
         end if;
      end loop;
      for K in N0 .. N-1 loop
	 M := K;
	 for I in K+1 .. N loop
	    if abs(A(I,K)) > abs(A(M,K)) then
	       M := I;
	    end if;
	 end loop;
	 Pvt(K) := M;
	 if M /= K then
	    Pvt(N) := -Pvt(N);
	 end if;
	 T := A(M,K);
	 A(M,K) := A(K,K);
	 A(K,K) := T;
	 if T /= Zero_Scalar then
	    for I in K+1 .. N loop
	       A(I,K) := -A(I,K)/T;
	    end loop;
	    for J in K+1 .. N loop
	       T := A(M,J);
	       A(M,J) := A(K,J);
	       A(K,J) := T;
	       if T /= Zero_Scalar then
		  for I in K+1 .. N loop
		     A(I,J) := A(I,J)+A(I,K)*T;
		  end loop;
	       end if;
	    end loop;
	 end if;
      end loop;
      for K in N0 .. N loop
	 T := Zero_Scalar;
	 for I in N0 .. K-1 loop
	    T := T+A(I,K)*Work(I);
	 end loop;
	 if A(K,K) = Zero_Scalar then
	    Message("Vectors.Decomp",Singular);
	 end if;
	 if T >= Zero_Scalar then
	    Work(K) := -(T+One_Scalar)/A(K,K);
	 else
	    Work(K) := -(T-One_Scalar)/A(K,K);
	 end if;
      end loop;
      for K in reverse N0 .. N-1 loop
	 T := Zero_Scalar;
	 for I in K+1 .. N loop
	    T := T+A(I,K)*Work(K);
	 end loop;
	 Work(K) := T;
	 M := Pvt(K);
	 if M /= K then
	    T := Work(M);
	    Work(M) := Work(K);
	    Work(K) := T;
	 end if;
      end loop;
      Ynorm := Zero_Scalar;
      for I in N0 .. N loop
	 Ynorm := Ynorm+abs(Work(I));
      end loop;
      Solve(A,Work,Pvt);
      Znorm := Zero_Scalar;
      for I in N0 .. N loop
	 Znorm := Znorm+abs(Work(I));
      end loop;
      Cond := Anorm*Znorm/Ynorm;
      if Cond < One_Scalar then
	 Cond := One_Scalar;
      end if;
   end Decomp;
   
   
   procedure Solve(A: in Matrix; B: in out Vector; Pvt: in Pivot) is
      --- solve A*B2=B1
      --- input:   A     triangularized matrix obtained from Decomp
      --- Pvt   pivot vector obtained from Decomp
      --- B  vector B1
      --- output:  B     vector B2
      N0,N,M: Integer;
      T: Scalar;
   begin
      N0 := B'First;
      N := B'Last;
      for K in N0 .. N-1 loop
	 M := Pvt(K);
	 T := B(M);
	 B(M) := B(K);
	 B(K) := T;
	 for I in K+1 .. N loop
	    B(I) := B(I)+A(I,K)*T;
	 end loop;
      end loop;
      for K in reverse N0+1 .. N loop
	 B(K) := B(K)/A(K,K);
	 T := -B(K);
	 for I in N0 .. K-1 loop
	    B(I) := B(I)+A(I,K)*T;
	 end loop;
      end loop;
      B(N0) := B(N0)/A(N0,N0);
   end Solve;
   
   
   procedure Balanc(A: in out Matrix) is
      -- translated from numerical recipes
      NearOne: constant Scalar := Scal(95)/Scal(100);
      Radix: constant Scalar := Two_Scalar;
      Done: Boolean;
      N0,N: Integer;
      S,R,G,F,C,Sqrdx: Scalar;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      Sqrdx := Radix*Radix;
      loop
         Done := True;
         for I in N0 .. N loop
            C := Zero_Scalar;
            R := Zero_Scalar;
            for J in N0 .. N loop
               if J /= I then
                  C := C+Abs(A(J,I));
                  R := R+Abs(A(I,J));
               end if;
            end loop;
            if C /= Zero_Scalar and then R /= Zero_Scalar then
               G := R/Radix;
               F := One_Scalar;
               S := C+R;
               while C < G loop
                  F := F*Radix;
                  C := C*Sqrdx;
               end loop;
               G := R*Radix;
               while C > G loop
                  F := F/Radix;
                  C := C/Sqrdx;
               end loop;
               if (C+R)/F < NearOne*S then
                  Done := False;
                  G := One_Scalar/F;
                  for J in N0 .. N loop A(I,J) := A(I,J)*G; end loop;
                  for J in N0 .. N loop A(J,I) := A(J,I)*F; end loop;
               end if;
            end if;
         end loop;
         exit when Done;
      end loop;
   end Balanc;


   procedure ElmHes(A: in out Matrix) is
      -- translated from numerical recipes
      N0,N,L: Integer;
      X,Y: Scalar;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      if N-N0>1 then
         for M in N0+1 .. N-1 loop
            X := Zero_Scalar;
            L := M;
            for J in M .. N loop
               if Abs(A(J,M-1)) > Abs(X) then
                  X := A(J,M-1);
                  L := J;
               end if;
            end loop;
            if L /= M then
               for J in M-1 .. N loop
                  Y := A(L,J);
                  A(L,J) := A(M,J);
                  A(M,J) := Y;
               end loop;
               for J in N0 .. N loop
                  Y := A(J,L);
                  A(J,L) := A(J,M);
                  A(J,M) := Y;
               end loop;
            end if;
            if X /= Zero_Scalar then
               for I  in M+1 .. N loop
                  Y := A(I,M-1);
                  if Y /= Zero_Scalar then
                     Y := Y/X;
                     A(I,M-1) := Y;
                     for J in M .. N loop A(I,J) := A(I,J)-Y*A(M,J); end loop;
                     for J in N0 .. N loop A(J,M) := A(J,M)+Y*A(J,I); end loop;
                  end if;
               end loop;
            end if;
         end loop;
      end if;
      for J  in A'Range(2) loop
         for I  in J+2..N loop
            A(I,J):=Zero_Scalar;
         end loop;
      end loop;
   end ElmHes;


   procedure Hqr(A: in out Matrix; Wr,Wi: out Vector) is
      -- translated from numerical recipes

      function Sign(R,S: Scalar) return Scalar is
      begin
         if S<Zero_Scalar then
            return -Abs(R);
         else
            return Abs(R);
         end if;
      end Sign;
      C1: constant Scalar := Scal(75)/Scal(100);
      C2: constant Scalar := Scal(-4375)/Scal(10000);
      FoundRoot: Boolean;
      N0,N,Nn,M,L,Its: Integer;
      Z,Y,X,W,V,U,T,S,R,Q,P,Anorm: Scalar;
   begin
      N0 := A'First(1);
      N := A'Last(1);
      Anorm := Abs(A(N0,N0));
      for I in N0+1 .. N loop
         for J in I-1 .. N loop
            Anorm := Anorm+Abs(A(I,J));
         end loop;
      end loop;
      Nn := N;
      T := Zero_Scalar;
      while Nn>=N0 loop
         FoundRoot := False;
         Its := 0;
         while not FoundRoot loop
            L := Nn;
            while L>N0 loop
               S := Abs(A(L-1,L-1))+Abs(A(L,L));
               if S=Zero_Scalar then
                  S := Anorm;
               end if;
               exit when (Abs(A(L,L-1))+S) = S;
               L := L-1;
            end loop;
            X := A(Nn,Nn);
            if L=Nn then
               FoundRoot := True;
               Wr(Nn) := X+T;
               Wi(Nn) := Zero_Scalar;
               Nn := Nn-1;
            else
               Y := A(Nn-1,Nn-1);
               W := A(Nn,Nn-1)*A(Nn-1,Nn);
               if L = Nn-1 then
                  FoundRoot := True;
                  P := (Y-X)/Scal(2);
                  Q := P*P+W;
                  Z := Sqrt(Abs(Q));
                  X := X+T;
                  if Q >= Zero_Scalar then
                     Z := P+Sign(Z,P);
                     Wr(Nn) := X+Z;
                     Wr(Nn-1) := Wr(Nn);
                     if Z /= Zero_Scalar then Wr(Nn) := X-W/Z; end if;
                     Wi(Nn) := Zero_Scalar;
                     Wi(Nn-1) := Zero_Scalar;
                  else
                     Wr(Nn) := X+P;
                     Wr(Nn-1) := Wr(Nn);
                     Wi(Nn) := Z;
                     Wi(Nn-1) := -Z;
                  end if;
                  Nn := Nn-2;
               else
                  --                  if Its=20 or Its=40 then
                  if Its=20 or (Its=40 or (Its=60 or Its=80)) then
                     T := T+X;
                     for I in N0 .. Nn loop
                        A(I,I) := A(I,I)-X;
                     end loop;
                     S := Abs(A(Nn,Nn-1))+Abs(A(Nn-1,Nn-2));
                     X := C1*S;
                     Y := X;
                     W := C2*S*S;
                  elsif Its =100 then
                     --                     Put("Too many iterations in Hqr");
                     --                     raise Getting_Tired;
                     Message("Vectors-Num.Eigenvalues",Giving_Up);
                  end if;
                  Its := Its+1;
                  M := Nn-2;
                  if M>=L then
                     loop
                        Z := A(M,M);
                        R := X-Z;
                        S := Y-Z;
                        P := (R*S-W)/A(M+1,M)+A(M,M+1);
                        Q := A(M+1,M+1)-Z-R-S;
                        R := A(M+2,M+1);
                        S := Abs(P)+Abs(Q)+Abs(R);
                        P := P/S;
                        Q := Q/S;
                        R := R/S;
                        exit when M=L;
                        U := Abs(A(M,M-1))*(Abs(Q)+Abs(R));
                        V := Abs(P)*(Abs(A(M-1,M-1))+Abs(Z)+Abs(A(M+1,M+1)));
                        exit when U+V=V;
                        M := M-1;
                     end loop;
                  end if;
                  for I in M+2 .. Nn loop
                     A(I,I-2) := Zero_Scalar;
                     if  I /= (M+2) then A(I,I-3) := Zero_Scalar; end if;
                  end loop;
                  for K in M .. Nn-1 loop
                     if K /= M then
                        P := A(K,K-1);
                        Q := A(K+1,K-1);
                        R := Zero_Scalar;
                        if K /= (Nn-1) then
                           R := A(K+2,K-1);
                        end if;
                        X := Abs(P)+Abs(Q)+Abs(R);
                        if X /= Zero_Scalar then
                           P := P/X;
                           Q := Q/X;
                           R := R/X;
                        end if;
                     end if;
                     S := Sign(Sqrt(P*P+Q*Q+R*R),P);
                     if S /= Zero_Scalar then
                        if K=M then
                           if L /= M then A(K,K-1) := -A(K,K-1); end if;
                        else
                           A(K,K-1) := -S*X;
                        end if;
                        P := P+S;
                        X := P/S;
                        Y := Q/S;
                        Z := R/S;
                        Q := Q/P;
                        R := R/P;
                        for J in K .. Nn loop
                           P := A(K,J)+Q*A(K+1,J);
                           if K /= (Nn-1) then
                              P := P+R*A(K+2,J);
                              A(K+2,J) := A(K+2,J)-P*Z;
                           end if;
                           A(K+1,J) := A(K+1,J)-P*Y;
                           A(K,J) := A(K,J)-P*X;
                        end loop;
                        for I in L .. Integer'Min(Nn,K+3) loop
                           P := X*A(I,K)+Y*A(I,K+1);
                           if K /= (Nn-1) then
                              P := P+Z*A(I,K+2);
                              A(I,K+2) := A(I,K+2)-P*R;
                           end if;
                           A(I,K+1) := A(I,K+1)-P*Q;
                           A(I,K) := A(I,K)-P;
                        end loop;
                     end if;
                  end loop;
               end if;
            end if;
         end loop;
      end loop;
   end Hqr;


   procedure Eigenvalues(A: in Matrix; Br,Bi: out Vector) is
      Aux: Matrix(A'Range(1),A'Range(2));
   begin
      Aux := A;
      Balanc(Aux);
      ElmHes(Aux);
      Hqr(Aux,Br,Bi);
      for J in Br'Range loop
         New_Line;
         Put(Current_Output,Br(J));Put("+I*");
         Put(Current_Output,Bi(J));
      end loop;

   end Eigenvalues;


   procedure Eigenvector(A: in Matrix; R: in Scalar; B: out Vector; Eps: in Scalar) is
      Iter: constant Integer := 200000;
      C,Epss: Scalar;
      Pvt: Pivot(B'Range);
      Bt: Vector(B'Range);
      Aux: Matrix(A'Range(1),A'Range(2));
   begin
      Epss:=Eps;
      
      while Epss<Scal(1000) loop
	 Aux := A;
	 C := One_Scalar/Sqrt(Scal(B'Last-B'First+1));
	 for I in B'Range loop
	    Aux(I,I) := Aux(I,I)-R;
	    B(I) := C;
	 end loop;
	 Decomp(Aux,C,Pvt,Bt);
	 for I in 1 .. Iter loop
	    Bt := B;
	    Solve(Aux,B,Pvt);
	    C := One_Scalar/Norm2(B);
	    if (B*Bt)<Zero_Scalar then C := -C; end if;
	    Mult(C,B);
	    Sub(B,Bt);
	    if Norm2(Bt)<Epss then return; end if;
	 end loop;
	 Epss:=Scal(10)*Epss;
      end loop;
      Message("Vectors-Num.Eigenvectors",Giving_Up);
      --raise Getting_Tired;
   end Eigenvector;
   

end Vectors.Num;
