
package body Vectors is


   function IsZero(V: Vector) return Boolean is
   begin
      for I in V'Range loop
	 if not IsZero(V(I)) then return False; end if;
      end loop;
      return True;
   end IsZero;

   procedure SetZero(V: out Vector) is
   begin
      for I in V'Range loop
	 V(I) := Zero_Scalar;
      end loop;
   end SetZero;

   procedure Neg(V: in out Vector) is
   begin
      for I in V'Range loop
	 Neg(V(I));
      end loop;
   end Neg;

   function "-"(V: Vector) return Vector is
      W: Vector(V'Range);
   begin
      for I in V'Range loop
	 W(I) := -V(I);
      end loop;
      return W;
   end "-";

   procedure Add(V1: in Vector; V2: in out Vector) is
   begin
      for I in V1'Range loop
	 Add(V1(I),V2(I));
      end loop;
   end Add;

   procedure Sum(V1,V2: in Vector; V3: out Vector) is
   begin
      for I in V1'Range loop
	 V3(I) := V1(I)+V2(I);
      end loop;
   end Sum;

   pragma Inline (Sum);

   function "+"(V1,V2: Vector) return Vector is
      V3: Vector(V1'Range);
   begin
      Sum(V1,V2,V3);
      return V3;
   end "+";

   procedure Sub(V1: in Vector; V2: in out Vector) is
   begin
      for I in V1'Range loop
	 Sub(V1(I),V2(I));
      end loop;
   end Sub;

   procedure Diff(V1,V2: in Vector; V3: out Vector) is
   begin
      for I in V1'Range loop
	 V3(I) := V1(I)-V2(I);
      end loop;
   end Diff;

   pragma Inline (Diff);

   function "-"(V1,V2: Vector) return Vector is
      V3: Vector(V1'Range);
   begin
      Diff(V1,V2,V3);
      return V3;
   end "-";

   procedure Mult(S: in Scalar; V: in out Vector) is
   begin
      for I in V'Range loop
	 Mult(S,V(I));
      end loop;
   end Mult;

   procedure Prod(S: in Scalar; V1: in Vector; V2: out Vector) is
   begin
      for I in V1'Range loop
	 V2(I) := S*V1(I);
      end loop;
   end Prod;

   pragma Inline (Prod);

   function "*"(S: Scalar; V: Vector) return Vector is
      W: Vector(V'Range);
   begin
      Prod(S,V,W);
      return W;
   end "*";

   function "*"(V1, V2: Vector) return Scalar is
      S: Scalar:=Zero_Scalar;
   begin
      for I in V1'Range loop
	 S:=S+V1(I)*V2(I);
      end loop;
      return S;
   end "*";

   procedure MultAdd(S: in Scalar; V1: in Vector; V2: in out Vector) is
   begin
      for I in V1'Range loop
	 MultAdd(S,V1(I),V2(I));
      end loop;
   end MultAdd;

   ------------

   function IsZero(A: Matrix) return Boolean is
   begin
      for I in A'Range(1) loop
	 for J in A'Range(2) loop
	    if not IsZero(A(I,J)) then return False; end if;
	 end loop;
      end loop;
      return True;
   end IsZero;

   procedure SetZero(A: out Matrix) is
   begin
      for I in A'Range(1) loop
	 for J in A'Range(2) loop
	    A(I,J) := Zero_Scalar;
	 end loop;
      end loop;
   end SetZero;

   procedure ZeroRow(I: in Integer; A: in out Matrix) is
   begin
      for J in A'Range(2) loop
	 A(I,J) := Zero_Scalar;
      end loop;
   end ZeroRow;

   procedure SetRow(I: in Integer; V: in Vector; A: in out Matrix) is
   begin
      for J in A'Range(2) loop
	 A(I,J) := V(J);
      end loop;
   end SetRow;

   function GetRow(I: Integer; A: Matrix) return Vector is
      V: Vector(A'Range(2));
   begin
      for J in A'Range(2) loop
	 V(J) := A(I,J);
      end loop;
      return V;
   end GetRow;

   procedure ZeroColumn(J: in Integer; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
	 A(I,J) := Zero_Scalar;
      end loop;
   end ZeroColumn;

   procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
	 A(I,J) := V(I);
      end loop;
   end SetColumn;

   function GetColumn(J: Integer; A: Matrix) return Vector is
      V: Vector(A'Range(1));
   begin
      for I in A'Range(1) loop
	 V(I) := A(I,J);
      end loop;
      return V;
   end GetColumn;

   function Transpose(A: Matrix) return Matrix is
      B: Matrix(A'Range(2),A'Range(1));
   begin
      for I in A'Range(1) loop
	 for J in A'Range(2) loop
	    B(J,I) := A(I,J);
	 end loop;
      end loop;
      return B;
   end Transpose;

   procedure Neg(A: in out Matrix) is
   begin
      for I in A'Range(1) loop
	 for J in A'Range(2) loop
	    Neg(A(I,J));
	 end loop;
      end loop;
   end Neg;

   function "-"(A: Matrix) return Matrix is
      B: Matrix(A'Range(1),A'Range(2));
   begin
      for I in A'Range(1) loop
	 for J in A'Range(2) loop
	    B(I,J) := -A(I,J);
	 end loop;
      end loop;
      return B;
   end "-";

   procedure Add(A1: in Matrix; A2: in out Matrix) is
   begin
      for I in A1'Range(1) loop
	 for J in A1'Range(2) loop
	    Add(A1(I,J),A2(I,J));
	 end loop;
      end loop;
   end Add;

   procedure Sum(A1,A2: in Matrix; A3: out Matrix) is
   begin
      for I in A1'Range(1) loop
	 for J in A1'Range(2) loop
	    A3(I,J) := A1(I,J)+A2(I,J);
	 end loop;
      end loop;
   end Sum;

   pragma Inline (Sum);

   function "+"(A1,A2: Matrix) return Matrix is
      A3: Matrix(A1'Range(1),A1'Range(2));
   begin
      Sum(A1,A2,A3);
      return A3;
   end "+";

   procedure Sub(A1: in Matrix; A2: in out Matrix) is
   begin
      for I in A1'Range(1) loop
	 for J in A1'Range(2) loop
	    Sub(A1(I,J),A2(I,J));
	 end loop;
      end loop;
   end Sub;

   procedure Diff(A1,A2: in Matrix; A3: out Matrix) is
   begin
      for I in A1'Range(1) loop
	 for J in A1'Range(2) loop
	    A3(I,J) := A1(I,J)-A2(I,J);
	 end loop;
      end loop;
   end Diff;

   pragma Inline (Sum);

   function "-"(A1,A2: Matrix) return Matrix is
      A3: Matrix(A1'Range(1),A1'Range(2));
   begin
      Diff(A1,A2,A3);
      return A3;
   end "-";

   procedure Mult(S: in Scalar; A: in out Matrix) is
   begin
      for I in A'Range(1) loop
	 for J in A'Range(2) loop
	    Mult(S,A(I,J));
	 end loop;
      end loop;
   end Mult;

   procedure Prod(S: in Scalar; A1: in Matrix; A2: out Matrix) is
   begin
      for I in A1'Range(1) loop
	 for J in A1'Range(2) loop
	    A2(I,J) := S*A1(I,J);
	 end loop;
      end loop;
   end Prod;

   pragma Inline (Prod);

   function "*"(S: Scalar; A: Matrix) return Matrix is
      B: Matrix(A'Range(1),A'Range(2));
   begin
      Prod(S,A,B);
      return B;
   end "*";

   procedure MultAdd(S: in Scalar; A1: in Matrix; A2: in out Matrix) is
   begin
      for I in A1'Range(1) loop
	 for J in A1'Range(2) loop
	    MultAdd(S,A1(I,J),A2(I,J));
	 end loop;
      end loop;
   end MultAdd;

   procedure Mult(A: in Matrix; V: in out Vector) is
      W: Vector := V;
   begin
      Prod(A,W,V);
   end Mult;

   procedure Prod(A: in Matrix; V1: in Vector; V2: out Vector) is
      S: Scalar;
   begin
      for I in V2'Range loop
	 S := Zero_Scalar;
	 for J in V1'Range loop
	    MultAdd(A(I,J),V1(J),S);
	 end loop;
	 V2(I) := S;
      end loop;
   end Prod;

   function "*"(A: Matrix; V: Vector) return Vector is
      W: Vector(A'Range(2));
   begin
      Prod(A,V,W);
      return W;
   end "*";

   procedure MultAdd(A: in Matrix; V1: in Vector; V2: in out Vector) is
      S: Scalar;
   begin
      for I in V2'Range loop
	 S := V2(I);
	 for J in V1'Range loop
	    MultAdd(A(I,J),V1(J),S);
	 end loop;
	 V2(I) := S;
      end loop;
   end MultAdd;

   procedure Mult(A1: in Matrix; A2: in out Matrix) is
      B: Matrix := A2;
   begin
      Prod(A1,B,A2);
   end Mult;


   procedure Mult(A1: in Matrix; A2: in out  Tensor) is
      S: Scalar;
      A3: Tensor(A1'Range(1),A2'Range(2),A2'Range(3));
   begin
      for I in A3'Range(1) loop
	 for M in A3'Range(2) loop
	    for K in A3'Range(3) loop
	       S := Zero_Scalar;
	       for J in A2'Range(2) loop
		  MultAdd(A1(I,J),A2(J,M,K),S);
	       end loop;
	       A3(I,M,K) := S;
	    end loop;
	 end loop;
      end loop;
      A2:=A3;
   end Mult;


   function "*"(A1: Matrix; A2: Tensor) return Tensor is
      A3: Tensor(A1'Range(1),A2'Range(2),A2'Range(3));
      S: Scalar;
   begin
      for I in A3'Range(1) loop
	 for M in A3'Range(2) loop
	    for K in A3'Range(3) loop
	       S := Zero_Scalar;
	       for J in A2'Range(2) loop
		  MultAdd(A1(I,J),A2(J,M,K),S);
	       end loop;
	       A3(I,M,K) := S;
	    end loop;
	 end loop;
      end loop;
      return A3;
   end "*";


   procedure Mult(A1: in out Tensor; A2,A3: in Matrix) is
      S: Scalar;
      A4: Tensor(A1'Range(1),A2'Range(2),A3'Range(2));
   begin
      for I in A4'Range(1) loop
	 for M in A4'Range(2) loop
	    for K in A4'Range(3) loop
	       S := Zero_Scalar;
	       for L in A2'Range(2) loop
		  for N in A3'Range(2) loop
		     S:=S+A1(I,L,N)*A2(L,M)*A3(N,K);
		  end loop;
	       end loop;
	       A4(I,M,K) := S;
	    end loop;
	 end loop;
      end loop;
      A1:=A4;
   end Mult;


   function Mult(A1: Tensor; A2,A3: Matrix) return Tensor is
      S: Scalar;
      A4: Tensor(A1'Range(1),A2'Range(2),A3'Range(2));
   begin
      for I in A4'Range(1) loop
	 for M in A4'Range(2) loop
	    for K in A4'Range(3) loop
	       S := Zero_Scalar;
	       for L in A2'Range(2) loop
		  for N in A3'Range(2) loop
		     S:=S+A1(I,L,N)*A2(L,M)*A3(N,K);
		  end loop;
	       end loop;
	       A4(I,M,K) := S;
	    end loop;
	 end loop;
      end loop;
      return A4;
   end Mult;


   procedure Add(A1: in Tensor; A2: in  out Tensor) is
   begin
      for I in A1'Range(1) loop
	 for M in A1'Range(2) loop
	    for K in A1'Range(3) loop
	       A2(I,M,K):=A1(I,M,K)+A2(I,M,K);
	    end loop;
	 end loop;
      end loop;
   end Add;


   function "+"(A1, A2: Tensor) return Tensor is
      A3: Tensor(A1'Range(1),A1'Range(2),A1'Range(3));
   begin
      for I in A1'Range(1) loop
	 for M in A1'Range(2) loop
	    for K in A1'Range(3) loop
	       A3(I,M,K):=A1(I,M,K)+A2(I,M,K);
	    end loop;
	 end loop;
      end loop;
      return A3;
   end "+";


   procedure Prod(A1,A2: in Matrix; A3: out Matrix) is
      S: Scalar;
   begin
      for I in A3'Range(1) loop
	 for J in A3'Range(2) loop
	    S := Zero_Scalar;
	    for K in A1'Range(2) loop
	       MultAdd(A1(I,K),A2(K,J),S);
	    end loop;
	    A3(I,J) := S;
	 end loop;
      end loop;
   end Prod;

   function "*"(A1,A2: Matrix) return Matrix is
      A3: Matrix(A1'Range(1),A2'Range(2));
   begin
      Prod(A1,A2,A3);
      return A3;
   end "*";

   procedure MultAdd(A1,A2: in Matrix; A3: in out Matrix) is
      S: Scalar;
   begin
      for I in A3'Range(1) loop
	 for J in A3'Range(2) loop
	    S := A3(I,J);
	    for K in A1'Range(2) loop
	       MultAdd(A1(I,K),A2(K,J),S);
	    end loop;
	    A3(I,J) := S;
	 end loop;
      end loop;
   end MultAdd;

   ------------

   procedure SetConst(S: in Scalar; P: out Polynom1) is
   begin
      SetZero(P);
      P(0) := S;
   end SetConst;

   function Val(X: Scalar; P: Polynom1) return Scalar is
      S: Scalar := Zero_Scalar;
   begin
      for I in reverse P'Range loop
	 S := X*S+P(I);
      end loop;
      return S;
   end Val;

   procedure Scale(S: in Scalar; P: in out Polynom1) is
      T: Scalar := S;
   begin
      for I in 1 .. P'Last loop
	 P(I) := T*P(I);
	 T := T*S;
      end loop;
   end Scale;

   procedure Scale(S: in Scalar; P1: in Polynom1; P2: out Polynom1) is
      T: Scalar := S;
   begin
      P2(0) := P1(0);
      for I in 1 .. P1'Last loop
	 P2(I) := T*P1(I);
	 T := T*S;
      end loop;
   end Scale;

   procedure Translate(S: in Scalar; P1: in Polynom1; P2: out Polynom1) is
   begin
      SetZero(P2);
      for I in P1'Range loop
	 for K in reverse 1 .. I loop
	    P2(K) := P2(K-1)-S*P2(K);
	 end loop;
	 P2(0) := P1(P1'Last-I)-S*P2(0);
      end loop;
   end Translate;

   procedure SetConst(S: in Scalar; P: in out Polynom2) is
   begin
      SetZero(P);
      P(0,0) := S;
   end SetConst;

   function Val(X,Y: Scalar; P: Polynom2) return Scalar is
      S,T: Scalar;
   begin
      S := Zero_Scalar;
      for I in reverse P'Range(1) loop
	 T := Zero_Scalar;
	 for J in reverse P'Range(2) loop
	    T := Y*T+P(I,J);
	 end loop;
	 S := X*S+T;
      end loop;
      return S;
   end Val;

   procedure Scale(S1,S2: in Scalar; P: in out Polynom2) is
      T1: Scalar := S1;
      T2: Scalar := S2;
   begin
      for J in 1 .. P'Last(2) loop
	 P(0,J) := T2*P(0,J);
	 T2 := T2*S2;
      end loop;
      for I in 1 .. P'Last(1) loop
	 T2 := S1;
	 for J in P'Range(2) loop
	    P(I,J) := T2*P(I,J);
	    T2 := T2*S2;
	 end loop;
	 T1 := T1*S1;
      end loop;
   end Scale;

   procedure Scale(S1,S2: in Scalar; P1: in Polynom2; P2: out Polynom2) is
      T1: Scalar := S1;
      T2: Scalar := S2;
   begin
      P2(0,0) := P1(0,0);
      for J in 1 .. P1'Last(2) loop
	 P2(0,J) := T2*P1(0,J);
	 T2 := T2*S2;
      end loop;
      for I in 1 .. P1'Last(1) loop
	 T2 := S1;
	 for J in P1'Range(2) loop
	    P2(I,J) := T2*P1(I,J);
	    T2 := T2*S2;
	 end loop;
	 T1 := T1*S1;
      end loop;
   end Scale;

   procedure Translate1(S: in Scalar; P1: in Polynom2; P2: out Polynom2) is
   begin
      for J in P1'Range(2) loop
	 ZeroColumn(J,P2);
	 for I in P1'Range(1) loop
	    for K in reverse 1 .. I loop
	       P2(K,J) := P2(K-1,J)-S*P2(K,J);
	    end loop;
	    P2(0,J) := P1(P1'Last(1)-I,J)-S*P2(0,J);
	 end loop;
      end loop;
   end Translate1;

   procedure Translate2(S: in Scalar; P1: in Polynom2; P2: out Polynom2) is
   begin
      for I in P1'Range(1) loop
	 ZeroRow(I,P2);
	 for J in P1'Range(2) loop
	    for K in reverse 1 .. J loop
	       P2(I,K) := P2(I,K-1)-S*P2(I,K);
	    end loop;
	    P2(I,0) := P1(I,P1'Last(2)-J)-S*P2(I,0);
	 end loop;
      end loop;
   end Translate2;

   procedure Translate(S1,S2: in Scalar; P1: in Polynom2; P2: out Polynom2) is
      Pt: Polynom2(P1'Range(1),P1'Range(2));
   begin
      Translate1(S1,P1,Pt);
      Translate2(S2,Pt,P2);
   end Translate;

end Vectors;
